from databricks.sdk import WorkspaceClient

from laktory._useragent import DATABRICKS_USER_AGENT
from laktory._useragent import VERSION
from laktory.models.stacks.stack import Stack
from laktory.dispatcher.pipelinerunner import PipelineRunner
from laktory.dispatcher.jobrunner import JobRunner


class Dispatcher:
    """
    The dispatcher is a manager that can be used to run and monitor remote jobs
    and pipelines defined in a stack. It is generally used through Laktory CLI
    `run` command, but may be used directly in scripts and python programs.

    Parameters
    ----------
    stack
        Stack object
    env
        Selected environment

    Examples
    --------
    ```py
    from laktory import models
    from laktory import Dispatcher

    with open("./stack.yaml") as fp:
        stack = models.Stack.model_validate_yaml(fp)

    dispatcher = Dispatcher(stack=stack)
    dispatcher.get_resource_ids()
    pl = dispatcher.resources["pl-stock-prices"]
    job = dispatcher.resources["job-stock-prices"]

    # Run pipeline
    pl.run(current_run_action="WAIT", full_refresh=False)

    # Run job
    job.run(current_run_action="CANCEL")
    ```
    """

    def __init__(self, stack: Stack = None, env: str = None):
        self.stack = stack
        self._env = env
        self._wc = None
        self.resources = {}

        self.init_resources()

    def init_resources(self):
        """Set resource for each of the resources defined in the stack"""
        for k, pl in self.stack.resources.pipelines.items():
            self.resources[k] = PipelineRunner(dispatcher=self, name=pl.name)

        for k, job in self.stack.resources.jobs.items():
            self.resources[k] = JobRunner(dispatcher=self, name=job.name)

    # ----------------------------------------------------------------------- #
    # Environment                                                             #
    # ----------------------------------------------------------------------- #

    @property
    def env(self) -> str:
        """Selected environment"""
        return self._env

    @env.setter
    def env(self, value):
        """Set environment"""
        self._env = value
        self._wc = None

    # ----------------------------------------------------------------------- #
    # Workspace Client                                                        #
    # ----------------------------------------------------------------------- #

    @property
    def _workspace_arguments(self):
        data = {}
        if self.stack.backend == "pulumi":
            config = self.stack.to_pulumi(env=self.env).model_dump()["config"]
            for k, v in config.items():
                if k.startswith("databricks"):
                    _k = k.split(":")[1]
                    data[_k] = v
        elif self.stack.backend == "terraform":
            providers = self.stack.to_terraform(env=self.env).model_dump()["provider"]
            for k in providers:
                if "databricks" in k.lower():
                    data = providers[k]
                    break

        kwargs = {}
        for k in [
            "host",
            "account_id",
            "username",
            "password",
            "client_id",
            "client_secret",
            "token",
            "profile",
            "config_file",
            "azure_workspace_resource_id",
            "azure_client_secret",
            "azure_client_id",
            "azure_tenant_id",
            "azure_environment",
            "auth_type",
            "cluster_id",
            "google_credentials",
            "google_service_account",
            "debug_truncate_bytes",
            "debug_headers",
            "product",
            "product_version",
        ]:
            if k in data:
                kwargs[k] = data[k]

        return kwargs

    @property
    def wc(self) -> WorkspaceClient:
        """Databricks Workspace Client"""
        if self._wc is None:
            self._wc = WorkspaceClient(**self._workspace_arguments)
            self._wc.config.with_user_agent_extra(
                key=DATABRICKS_USER_AGENT,
                value=VERSION,
            )
        return self._wc

    # ----------------------------------------------------------------------- #
    # Resources                                                               #
    # ----------------------------------------------------------------------- #

    def get_resource_ids(self, env=None):
        """
        Get resource ids for each of the resources defined in the stack in the
        provided environment `env`.
        """
        if env is not None:
            self.env = env

        for r in self.resources.values():
            r.get_id()

    # ----------------------------------------------------------------------- #
    # Run                                                                     #
    # ----------------------------------------------------------------------- #

    def run_job(self, job_name: str, *args, **kwargs):
        """
        Run job with name `job_name`

        Parameters
        ----------
        job_name:
            Name of the job
        *args:
            Arguments passed to `JobRunner.run()`
        **kwargs:
            Keyword arguments passed to `JobRunner.run()`
        """
        job = self.resources[job_name]
        job.run(*args, **kwargs)

    def run_pipeline(self, pipeline_name: str, *args, **kwargs):
        """
        Run pipeline with name `pipeline_name`

        Parameters
        ----------
        pipeline_name:
            Name of the pipeline
        *args:
            Arguments passed to `JobRunner.run()`
        **kwargs:
            Keyword arguments passed to `JobRunner.run()`
        """
        pl = self.resources[pipeline_name]
        pl.run(*args, **kwargs)



import logging
import sys

from ._settings import settings

# --------------------------------------------------------------------------- #
# Get logger                                                                  #
# --------------------------------------------------------------------------- #

# Stream Handler
stream_handler = logging.StreamHandler(sys.stdout)
format = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
stream_handler.setFormatter(format)


def get_logger(name, stream=True):
    # Change logger class
    # TODO: Place Holder if we need to customize logger class
    # logging.setLoggerClass(LaktoryLogger)

    # Get logger
    logger = logging.getLogger(name)

    # Set level
    logger.setLevel(settings.log_level)

    # Set options
    _options = {
        "stream": stream,
    }

    # Set states
    _initialized = hasattr(logger, "_initialized")
    _options_changed = False
    if _initialized:
        _options_changed = logger._options != _options

    # Update handlers
    if not _initialized or _options_changed:
        handlers = list(logger.handlers)
        for handler in handlers:
            logger.removeHandler(handler)

        if stream:
            logger.addHandler(stream_handler)

    # Reset logger class
    logging.setLoggerClass(logging.Logger)

    # Update state
    logger._initialized = True
    logger._options = _options

    return logger



from typing import Any
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field

from databricks.sdk import WorkspaceClient


class DispatcherRunner(BaseModel):
    """
    Base runner for jobs, pipelines and other executables.

    Attributes
    ----------
    name:
        Name of the resource attached to the runner
    id:
        ID of the deployed resource
    dispatcher:
        Dispatcher managing the runs
    """

    model_config = ConfigDict(extra="forbid")
    name: str = None
    id: str = None
    dispatcher: Any = Field(default=None, exclude=True)

    @property
    def wc(self) -> WorkspaceClient:
        """Databricks Workspace Client"""
        return self.dispatcher.wc

    def get_id(self) -> str:
        raise NotImplementedError()

    def run(self, wait=True):
        raise NotImplementedError()



from typing import Union
from pydantic import Field
from pydantic import ConfigDict

from laktory._settings import settings
from laktory.models.basemodel import BaseModel
from laktory.models.dataproducer import DataProducer


class DataEventHeader(BaseModel):
    """
    Data Event Header class defines the context (metadata) describing a data
    event. It is generally used to read data from a storage location or to
    build a `EventDataSource`.

    Attributes
    ----------
    name
        Data event name
    description
        Data event description
    producer
        Data event producer
    event_root
        Root path for specific event. Default value: `{settings.workspace_landing_root}/events/my-event`
    events_root
        Root path for all events. Default value: `{settings.workspace_landing_root}/events/`

    Examples
    ---------
    ```python
    from laktory import models

    event = models.DataEventHeader(
        name="stock_price",
        producer={"name": "yahoo-finance"},
    )
    print(event)
    '''
    variables={} name='stock_price' description=None producer=DataProducer(variables={}, name='yahoo-finance', description=None, party=1) events_root_=None event_root_=None
    '''

    print(event.event_root)
    #> /Volumes/dev/sources/landing/events/yahoo-finance/stock_price/
    ```
    """

    model_config = ConfigDict(extra="forbid", populate_by_name=True)
    name: str = Field(...)
    description: Union[str, None] = Field(default=None)
    producer: DataProducer = Field(default=None)
    events_root_: Union[str, None] = Field(None, alias="events_root")
    event_root_: Union[str, None] = Field(None, alias="event_root")

    # ----------------------------------------------------------------------- #
    # Paths                                                                   #
    # ----------------------------------------------------------------------- #

    @property
    def events_root(self) -> str:
        """Must be computed to dynamically account for settings (env variable at run time)"""
        if self.events_root_:
            return self.events_root_
        return settings.workspace_landing_root + "events/"

    @property
    def event_root(self) -> str:
        """
        Root path for the event. Default path is `{self.events_roots}/{producer_name}/{event_name}/`, but it may
        be overwritten by self.event_root_.

        Returns
        -------
        str
            Event path
        """
        if self.event_root_:
            return self.event_root_

        producer = ""
        if self.producer is not None:
            producer = self.producer.name + "/"
        v = f"{self.events_root}{producer}{self.name}/"
        return v



from pyspark.sql import SparkSession
import os
import pandas as pd

from laktory.models.datasources import EventDataSource
from laktory.models.datasources import TableDataSource
from laktory._testing import Paths
from laktory._testing import spark

paths = Paths(__file__)

# DataFrame
pdf = pd.DataFrame(
    {
        "x": [1, 2, 3],
        "a": [1, -1, 1],
        "b": [2, 0, 2],
        "c": [3, 0, 3],
        "n": [4, 0, 4],
    },
)
df0 = spark.createDataFrame(pdf)


def test_event_data_source():
    source = EventDataSource(
        name="stock_price",
        producer={"name": "yahoo_finance"},
    )
    assert (
        source.event_root
        == "/Volumes/dev/sources/landing/events/yahoo_finance/stock_price/"
    )
    assert not source.is_cdc


def test_event_data_source_read():
    source = EventDataSource(
        name="stock_price",
        producer={"name": "yahoo-finance"},
        events_root=os.path.join(paths.data, "./events/"),
        read_as_stream=False,
    )
    df = source.read(spark)
    assert df.count() == 80
    assert df.columns == [
        "data",
        "description",
        "event_root_",
        "name",
        "producer",
    ]


def test_table_data_source(df0=df0):
    source = TableDataSource(
        mock_df=df0,
        path="/Volumes/tables/stock_prices/",
        filter="b != 0",
        selects=["a", "b", "c"],
        renames={"a": "aa", "b": "bb", "c": "cc"},
        broadcast=True,
        spark_chain={
            "nodes": [
                {
                    "column": {"name": "chain"},
                    "spark_func_name": "lit",
                    "spark_func_args": ["chain"],
                }
            ]
        },
    )

    # Test paths
    assert source.path == "/Volumes/tables/stock_prices/"
    assert source.from_path
    assert source.path_or_full_name == "/Volumes/tables/stock_prices/"

    # Test reader
    df = source.read(spark)
    assert df.columns == ["aa", "bb", "cc", "chain"]
    assert df.count() == 2
    assert df.toPandas()["chain"].tolist() == ["chain", "chain"]

    # Select with rename
    source = TableDataSource(
        mock_df=df0,
        path="/Volumes/tables/stock_prices/",
        selects={"x": "x1", "n": "n1"},
    )
    df = source.read(spark)
    assert df.columns == ["x1", "n1"]
    assert df.count() == 3

    # Drop
    source = TableDataSource(
        mock_df=df0,
        path="/Volumes/tables/stock_prices/",
        drops=["b", "c", "n"],
    )
    df = source.read(spark)
    assert df.columns == ["x", "a"]
    assert df.count() == 3


if __name__ == "__main__":
    test_event_data_source()
    test_event_data_source_read()
    test_table_data_source()



from pydantic import BaseModel
from typing import Union
from pyspark.sql.dataframe import DataFrame

from laktory._logger import get_logger


logger = get_logger(__name__)


class OrderBy(BaseModel):
    sql_expression: str
    desc: bool = False


def window_filter(
    df,
    partition_by: Union[list[str], None],
    order_by: Union[list[OrderBy], None] = None,
    drop_row_index: bool = True,
    row_index_name: str = "_row_index",
    rows_to_keep: int = 1,
) -> DataFrame:
    """
    Apply spark window-based filtering

    Parameters
    ----------
    df:
        DataFrame
    partition_by
        Defines the columns used for grouping into Windows
    order_by:
        Defines the column used for sorting before dropping rows
    drop_row_index:
        If `True`, the row index column is dropped
    row_index_name:
        Group-specific and sorted rows index name
    rows_to_keep:
        How many rows to keep per window


    Examples
    --------
    ```py
    import pyspark.sql.functions as F

    from laktory._testing import spark

    df0 = spark.createDataFrame(
        [
            ["2023-01-01T00:00:00Z", "APPL", 200.0],
            ["2023-01-02T00:00:00Z", "APPL", 202.0],
            ["2023-01-03T00:00:00Z", "APPL", 201.5],
            ["2023-01-01T00:00:00Z", "GOOL", 200.0],
            ["2023-01-02T00:00:00Z", "GOOL", 202.0],
            ["2023-01-03T00:00:00Z", "GOOL", 201.5],
        ],
        ["created_at", "symbol", "price"],
    ).withColumn("created_at", F.col("created_at").cast("timestamp"))

    df = df0.window_filter(
        partition_by=["symbol"],
        order_by=[
            {"sql_expression": "created_at", "desc": True},
        ],
        drop_row_index=False,
        rows_to_keep=1,
    )

    print(df.toPandas())
    '''
      created_at symbol  price  _row_index
    0 2023-01-03   APPL  201.5           1
    1 2023-01-03   GOOL  201.5           1
    '''
    ```

    References
    ----------

    * [pyspark window](https://spark.apache.org/docs/3.1.3/api/python/reference/api/pyspark.sql.functions.window.html)
    """
    from pyspark.sql import Window
    import pyspark.sql.functions as F

    # Partition by
    w = Window.partitionBy(*partition_by)

    # Order by
    if order_by:
        order_bys = []
        for o in order_by:
            if not isinstance(o, OrderBy):
                o = OrderBy(**o)
            order_bys += [F.expr(o.sql_expression)]
            if o.desc:
                order_bys[-1] = order_bys[-1].desc()
        w = w.orderBy(*order_bys)

    # Set rows index
    df = df.withColumn(row_index_name, F.row_number().over(w))

    # Filter
    df = df.filter(F.col(row_index_name) <= rows_to_keep)
    if drop_row_index:
        df = df.drop(row_index_name)

    return df


if __name__ == "__main__":
    import pyspark.sql.functions as F

    from laktory._testing import spark

    df0 = spark.createDataFrame(
        [
            ["2023-01-01T00:00:00Z", "APPL", 200.0],
            ["2023-01-02T00:00:00Z", "APPL", 202.0],
            ["2023-01-03T00:00:00Z", "APPL", 201.5],
            ["2023-01-01T00:00:00Z", "GOOL", 200.0],
            ["2023-01-02T00:00:00Z", "GOOL", 202.0],
            ["2023-01-03T00:00:00Z", "GOOL", 201.5],
        ],
        ["created_at", "symbol", "price"],
    ).withColumn("created_at", F.col("created_at").cast("timestamp"))

    df = df0.window_filter(
        partition_by=["symbol"],
        order_by=[
            {"sql_expression": "created_at", "desc": True},
        ],
        drop_row_index=False,
        rows_to_keep=1,
    )

    print(df.toPandas())



from typing import Union
from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class ExternalLocationEncryptionDetailsSseEncryptionDetails(BaseModel):
    """
    Attributes
    ----------
    algorith:
    aws_kms_key_arn:
    """

    algorith: str = None
    aws_kms_key_arn: str = None


class ExternalLocationEncryptionDetails(BaseModel):
    """
    Attributes
    ----------
    sse_encryption_details:
    """

    sse_encryption_details: ExternalLocationEncryptionDetailsSseEncryptionDetails = None


class ExternalLocation(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks External Location

    Attributes
    ----------
    access_point:
        The ARN of the s3 access point to use with the external location (AWS).
    comment:
        User-supplied free-form text.
    credential_name:
        Name of the databricks.StorageCredential to use with this external location.
    encryption_details:
        The options for Server-Side Encryption to be used by each Databricks s3 client when connecting to S3 cloud
        storage (AWS).
    force_destroy:
        Destroy external location regardless of its dependents.
    force_update:
        Update external location regardless of its dependents.
    metastore_id:
        Metastore ID
    name:
        Name of External Location, which must be unique within the databricks_metastore. Change forces creation of a new
        resource.
    owner:
        Username/groupname/sp application_id of the external location owner.
    read_only:
        Indicates whether the external location is read-only.
    skip_validation:
        Suppress validation errors if any & force save the external location
    url:
        Path URL in cloud storage, of the form: s3://[bucket-host]/[bucket-dir] (AWS), abfss://[user]@[host]/[path]
        (Azure), gs://[bucket-host]/[bucket-dir] (GCP).

    Examples
    --------
    ```py
    ```
    """

    access_point: str = None
    comment: str = None
    credential_name: str = None
    encryption_details: ExternalLocationEncryptionDetails = None
    force_destroy: bool = None
    force_update: bool = None
    metastore_id: str = None
    name: str = None
    owner: str = None
    read_only: bool = None
    skip_validation: bool = None
    url: str = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:ExternalLocation"

    @property
    def pulumi_cls(self):
        import pulumi_databricks as databricks

        return databricks.ExternalLocation

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_external_location"



from typing import Union
from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource
from laktory.models.databricks.mwspermissionassignment import MwsPermissionAssignment


class Group(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks group

    Attributes
    ----------
    allow_cluster_create:
        When `True`, the group is allowed to have cluster create permissions
    display_name:
        Display name for the group.
    id:
        Id of the group. Generally used when the group is externally managed
        with an identity provider such as Azure AD, Okta or OneLogin.
    workspace_access
        When `True`, the group is allowed to have workspace access

    Examples
    --------
    ```py
    from laktory import models

    d = models.Group(display_name="role-engineers")
    ```
    """

    allow_cluster_create: bool = False
    display_name: str
    id: Union[str, None] = None
    workspace_access: bool = None
    workspace_permission_assignments: list[MwsPermissionAssignment] = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        """display name"""
        return self.display_name

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Group"

    @property
    def pulumi_cls(self):
        import pulumi_databricks as databricks

        return databricks.Group

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["workspace_permission_assignments"]

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - workspace permission assignments
        """
        resources = []
        if self.workspace_permission_assignments:
            for a in self.workspace_permission_assignments:
                if a.principal_id is None:
                    a.principal_id = f"${{resources.{self.resource_name}.id}}"
                resources += [a]
        return resources

    # TODO:
    # if group.id is None:
    #     self.group = databricks.Group(name, opts=opts, **group.model_pulumi_dump())
    #     group.id = self.group.id
    # else:
    #     self.group = databricks.Group.get(name, id=group.id)

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_group"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from typing import Union

from laktory.models.basemodel import BaseModel
from laktory.models.databricks.grants import Grants
from laktory.models.grants.schemagrant import SchemaGrant
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource
from laktory.models.sql.table import Table
from laktory.models.sql.volume import Volume


class Schema(BaseModel, PulumiResource, TerraformResource):
    """
    A schema (also called a database) is the second layer of Unity Catalog’s
    three-level namespace. A schema organizes tables and views.

    Attributes
    ----------
    catalog_name:
        Name of the catalog storing the schema
    comment:
        Text description of the catalog
    force_destroy:
        If `True` catalog can be deleted, even when not empty
    grants:
        List of grants operating on the schema
    isolation_mode:
        Whether the catalog is accessible from all workspaces or a specific set
        of workspaces. Can be ISOLATED or OPEN. Setting the catalog to ISOLATED
        will automatically allow access from the current workspace.
    name:
        Name of the catalog
    tables:
        List of tables stored in the schema
    volumes:
        List of volumes stored in the schema

    Examples
    --------
    ```py
    from laktory import models

    schema = models.Schema(
        catalog_name="dev",
        name="engineering",
        grants=[{"principal": "domain-engineering", "privileges": ["SELECT"]}],
    )
    ```

    References
    ----------

    * [Databricks Unity Schema](https://docs.databricks.com/en/data-governance/unity-catalog/index.html#schemas)
    * [Pulumi Databricks Schema](https://www.pulumi.com/registry/packages/databricks/api-docs/schema/)
    """

    catalog_name: Union[str, None] = None
    comment: Union[str, None] = None
    force_destroy: bool = True
    grants: list[SchemaGrant] = None
    name: str
    tables: list[Table] = []
    volumes: list[Volume] = []

    def model_post_init(self, __context):
        super().model_post_init(__context)
        for table in self.tables:
            table.catalog_name = self.catalog_name
            table.schema_name = self.name
        for volume in self.volumes:
            volume.catalog_name = self.catalog_name
            volume.schema_name = self.name

    # ----------------------------------------------------------------------- #
    # Computed fields                                                         #
    # ----------------------------------------------------------------------- #

    @property
    def parent_full_name(self) -> str:
        """Catalog full name `{catalog_name}`"""
        return self.catalog_name

    @property
    def full_name(self) -> str:
        """Schema full name `{catalog_name}.{schema_name}`"""
        _id = self.name
        if self.parent_full_name is not None:
            _id = f"{self.parent_full_name}.{_id}"
        return _id

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        """Schema full name (catalog.schema)"""
        return self.full_name

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - schema grants
        - tables
        - volumes
        """
        resources = []

        # Schema grants
        if self.grants:
            resources += [
                Grants(
                    resource_name=f"grants-{self.resource_name}",
                    schema=self.full_name,
                    grants=[
                        {"principal": g.principal, "privileges": g.privileges}
                        for g in self.grants
                    ],
                )
            ]

        if self.volumes:
            for v in self.volumes:
                resources += v.core_resources

        if self.tables:
            for t in self.tables:
                resources += t.core_resources

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Schema"

    @property
    def pulumi_cls(self):
        import pulumi_databricks as databricks

        return databricks.Schema

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["tables", "volumes", "grants"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_schema"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from laktory.dispatcher.dispatcher import Dispatcher
from laktory.dispatcher.jobrunner import JobRunner
from laktory.dispatcher.pipelinerunner import PipelineRunner
from laktory.dispatcher.dispatcherrunner import DispatcherRunner



import os
from pyspark.sql import SparkSession
from pyspark.sql import types as T
from pyspark.sql import functions as F
from pandas import Timestamp

import laktory
from laktory._testing.stockprices import spark
from laktory._testing.stockprices import df_slv
from laktory._testing.stockprices import df_meta

schema = T.StructType(
    [
        T.StructField("x@x", T.IntegerType()),
        T.StructField("y", T.ArrayType(T.StringType())),
        T.StructField(
            "z",
            T.StructType(
                [
                    T.StructField("id", T.IntegerType()),
                    T.StructField("email", T.StringType()),
                ]
            ),
        ),
        T.StructField(
            "u",
            T.ArrayType(
                T.StructType(
                    [
                        T.StructField("a", T.IntegerType()),
                        T.StructField("b", T.IntegerType()),
                    ]
                )
            ),
        ),
        T.StructField(
            "u2",
            T.ArrayType(
                T.StructType(
                    [
                        T.StructField("a", T.IntegerType()),
                        T.StructField("b", T.IntegerType()),
                    ]
                )
            ),
        ),
    ]
)

data = [
    (
        1,
        ["a", "b"],
        {"id": 3, "email": "@gmail.com"},
        [{"a": 1, "b": 2}, {"a": 1, "b": 2}],
        [{"a": 1, "b": 2}, {"a": 1, "b": 2}],
    ),
    (
        2,
        ["b", "c"],
        {"id": 2, "email": "@gmail.com"},
        [{"a": 1, "b": 2}, {"a": 1, "b": 2}],
        [{"a": 1, "b": 2}, {"a": 1, "b": 2}],
    ),
    (
        3,
        ["c", "d"],
        {"id": 1, "email": "@gmail.com"},
        [{"a": 1, "b": 2}, {"a": 1, "b": 2}],
        [{"a": 1, "b": 2}, {"a": 1, "b": 2}],
    ),
]
df = spark.createDataFrame(data, schema=schema)

spark = SparkSession.builder.appName("UnitTesting").getOrCreate()
spark.conf.set("spark.sql.session.timeZone", "UTC")


def test_df_schema_flat():
    df.show()
    df.printSchema()
    schema = df.schema_flat()
    assert schema == [
        "x@x",
        "y",
        "z",
        "z.id",
        "z.email",
        "u",
        "u[*].a",
        "u[*].b",
        "u2",
        "u2[*].a",
        "u2[*].b",
    ]


def test_df_has_column():
    assert df.has_column("x@x")
    assert df.has_column("`x@x`")
    assert df.has_column("u[0].a")


def test_watermark():
    df = df_slv.withWatermark("created_at", "1 hour")

    wm = df.watermark()

    assert wm.column == "created_at"
    assert wm.threshold == "1 hours"

    # TODO: Add test for spark connect DataFrame


def test_join():
    left = df_slv.filter(F.col("created_at") == "2023-09-01T00:00:00Z").filter(
        F.col("symbol") != "GOOGL"
    )
    other = df_meta.withColumnRenamed("symbol2", "symbol")

    other.show()

    df = left.smart_join(
        other=other,
        on=["symbol"],
        how="left",
    )

    # Test join columns uniqueness
    _df = df.withColumn("symbol2", F.lit("a"))
    _df = df.withColumn("symbol", F.lit("a"))
    _df = _df.select("symbol")

    # Test data
    data = df.toPandas().fillna(-1).to_dict(orient="records")
    print(data)
    assert data == [
        {
            "created_at": Timestamp("2023-09-01 00:00:00"),
            "open": 189.49000549316406,
            "close": 189.4600067138672,
            "currency": "USD",
            "first_traded": "1980-12-12T14:30:00.000Z",
            "symbol": "AAPL",
        },
        {
            "created_at": Timestamp("2023-09-01 00:00:00"),
            "open": 139.4600067138672,
            "close": 138.1199951171875,
            "currency": "USD",
            "first_traded": "1997-05-15T13:30:00.000Z",
            "symbol": "AMZN",
        },
        # {
        #     "created_at": Timestamp("2023-09-01 00:00:00"),
        #     "open": 137.4600067138672,
        #     "close": 135.66000366210938,
        #     "currency": "USD",
        #     "first_traded": "2004-08-19T13:30:00.00Z",
        #     "symbol": "GOOGL",
        # },
        {
            "created_at": Timestamp("2023-09-01 00:00:00"),
            "open": 331.30999755859375,
            "close": 328.6600036621094,
            "currency": -1,
            "first_traded": -1,
            "symbol": "MSFT",
        },
    ]

    # Join expression
    df2 = left.smart_join(
        other=other,
        on_expression="left.symbol == other.symbol",
    )
    assert df2.toPandas().equals(df.toPandas())


def test_join_outer():
    left = df_slv.filter(F.col("created_at") == "2023-09-01T00:00:00Z").filter(
        F.col("symbol") != "GOOGL"
    )
    other = df_meta.withColumnRenamed("symbol2", "symbol")

    df = left.smart_join(
        other=other,
        on=["symbol"],
        how="full_outer",
    )

    # Test join columns uniqueness
    _df = df.withColumn("symbol2", F.lit("a"))
    _df = df.withColumn("symbol", F.lit("a"))
    _df = _df.select("symbol")

    # Test data
    data = df.toPandas().fillna(-1).to_dict(orient="records")
    print(data)
    assert data == [
        {
            "created_at": Timestamp("2023-09-01 00:00:00"),
            "open": 189.49000549316406,
            "close": 189.4600067138672,
            "currency": "USD",
            "first_traded": "1980-12-12T14:30:00.000Z",
            "symbol": "AAPL",
        },
        {
            "created_at": Timestamp("2023-09-01 00:00:00"),
            "open": 139.4600067138672,
            "close": 138.1199951171875,
            "currency": "USD",
            "first_traded": "1997-05-15T13:30:00.000Z",
            "symbol": "AMZN",
        },
        {
            "created_at": -1,
            "open": -1,
            "close": -1,
            "currency": "USD",
            "first_traded": "2004-08-19T13:30:00.00Z",
            "symbol": "GOOGL",
        },
        {
            "created_at": Timestamp("2023-09-01 00:00:00"),
            "open": 331.30999755859375,
            "close": 328.6600036621094,
            "currency": -1,
            "first_traded": -1,
            "symbol": "MSFT",
        },
    ]


def test_join_watermark():
    # TODO
    pass


def test_aggregation():
    _df = df_slv.filter(F.col("created_at") < "2023-09-07T00:00:00Z")
    _df.show()

    # Window
    df = _df.groupby_and_agg(
        groupby_window={
            "time_column": "created_at",
            "window_duration": "1 day",
        },
        agg_expressions=[
            {
                "column": {"name": "min_open"},
                "spark_func_name": "min",
                "spark_func_args": ["open"],
            },
            {
                "column": {"name": "max_open"},
                "sql_expression": "max(open)",
            },
        ],
    ).sort("window")
    pdf = df.toPandas()
    assert "window" in df.columns
    assert pdf["min_open"].round(2).to_list() == [137.46, 135.44, 136.02]
    assert pdf["max_open"].round(2).to_list() == [331.31, 329.0, 333.38]

    # Symbol
    df = _df.groupby_and_agg(
        groupby_columns=["symbol"],
        agg_expressions=[
            {
                "column": {"name": "mean_close"},
                "sql_expression": "mean(close)",
            },
        ],
    ).sort("symbol")
    pdf = df.toPandas()
    assert "symbol" in df.columns
    assert pdf["mean_close"].round(2).to_list() == [187.36, 136.92, 135.3, 331.7]

    # Symbol and window
    df = _df.groupby_and_agg(
        groupby_window={
            "time_column": "created_at",
            "window_duration": "1 day",
        },
        groupby_columns=["symbol"],
        agg_expressions=[
            {
                "column": {"name": "count"},
                "sql_expression": "count(close)",
            },
        ],
    ).sort("symbol", "window")
    pdf = df.toPandas()
    assert "symbol" in df.columns
    assert "window" in df.columns
    assert (
        pdf["count"].tolist()
        == [
            1,
        ]
        * _df.count()
    )


def test_window_filter():
    df = df_slv.window_filter(
        partition_by=["symbol"],
        order_by=[
            {"sql_expression": "created_at", "desc": True},
        ],
        drop_row_index=False,
        rows_to_keep=2,
    ).select("created_at", "symbol", "_row_index")

    data = df.toPandas().to_dict(orient="records")
    assert data == [
        {
            "created_at": Timestamp("2023-09-29 00:00:00"),
            "symbol": "AAPL",
            "_row_index": 1,
        },
        {
            "created_at": Timestamp("2023-09-28 00:00:00"),
            "symbol": "AAPL",
            "_row_index": 2,
        },
        {
            "created_at": Timestamp("2023-09-29 00:00:00"),
            "symbol": "AMZN",
            "_row_index": 1,
        },
        {
            "created_at": Timestamp("2023-09-28 00:00:00"),
            "symbol": "AMZN",
            "_row_index": 2,
        },
        {
            "created_at": Timestamp("2023-09-29 00:00:00"),
            "symbol": "GOOGL",
            "_row_index": 1,
        },
        {
            "created_at": Timestamp("2023-09-28 00:00:00"),
            "symbol": "GOOGL",
            "_row_index": 2,
        },
        {
            "created_at": Timestamp("2023-09-29 00:00:00"),
            "symbol": "MSFT",
            "_row_index": 1,
        },
        {
            "created_at": Timestamp("2023-09-28 00:00:00"),
            "symbol": "MSFT",
            "_row_index": 2,
        },
    ]


if __name__ == "__main__":
    test_df_schema_flat()
    test_df_has_column()
    test_watermark()
    test_join()
    test_join_outer()
    test_join_watermark()
    test_aggregation()
    test_window_filter()



from laktory.models import Directory

directory = Directory(path=".laktory/pipelines/")


def test_directory():
    assert directory.path == ".laktory/pipelines/"
    assert directory.resource_name == "directory-laktory-pipelines"


if __name__ == "__main__":
    test_directory()



import os

script_root = os.path.dirname(__file__)
package_name = "laktory"
package_root = f"./{package_name}/"


def main():
    # Filepaths
    version_filepath = os.path.join(package_root, "_version.py")
    changelog_filepath = os.path.join("./", "CHANGELOG.md")
    body_filepath = os.path.join(script_root, "release_body.md")

    # Read version file
    with open(version_filepath) as fp:
        release_version = fp.read().split("=")[-1].strip().replace('"', "")
    print(f"Get changes for {package_name} {release_version}")

    # Read changelog
    with open(changelog_filepath, "r") as fp:
        content = fp.read()
    blocks = content.split("## [")

    # Select latest changes
    changes = blocks[1]
    changes = "\n".join(changes.split("\n")[1:]).strip()
    changes = changes.replace("###", "####")

    # Add header
    changes = f"### What’s new\n{changes}"

    # Add compare to previous release
    previous_version = blocks[2].split("]")[0]
    changes += f"\n\nFull [Changelog](https://github.com/okube-ai/{package_name}/compare/v{previous_version}...v{release_version}/)"

    print("Release Content")
    print("---------------")
    print(changes)
    print("--------------")

    # Write body
    print(f"Writing body {body_filepath}")
    with open(body_filepath, "w") as fp:
        fp.write(changes)

    if os.getenv("GITHUB_OUTPUT") is None:
        os.remove(body_filepath)


if __name__ == "__main__":
    main()



from typing import Any
from typing import Union
from typing import Literal

from pydantic import model_validator

from laktory._logger import get_logger
from laktory._settings import settings
from laktory.models.basemodel import BaseModel
from laktory.models.databricks.grants import Grants
from laktory.models.grants.tablegrant import TableGrant
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource
from laktory.models.sql.column import Column
from laktory.models.sql.tablebuilder import TableBuilder
from laktory.models.sql.tableexpectation import TableExpectation
from laktory.models.datasources.tabledatasource import TableDataSource

logger = get_logger(__name__)


class Table(BaseModel, PulumiResource, TerraformResource):
    """
    A table resides in the third layer of Unity Catalog’s three-level
    namespace. It contains rows of data. Laktory provides the mechanism to
    build the table data in the context of a data pipeline using the
    `builder` attribute.

    Attributes
    ----------
    builder:
        Instructions on how to build data from a source in the context of a
        data pipeline.
    catalog_name:
        Name of the catalog storing the table
    columns:
        List of columns stored in the table
    comment:
        Text description of the catalog
    data:
        Data to be used to populate the rows
    expectations:
        List of expectations for the table. Can be used as warnings, drop
        invalid records or fail a pipeline.
    grants:
        List of grants operating on the schema
    name:
        Name of the table
    primary_key:
        Name of the column storing a unique identifier for each row. It is used
        by the builder to drop duplicated rows.
    schema_name:
        Name of the schema storing the table
    table_type:
        Distinguishes a view vs. managed/external Table.
    timestamp_key:
        Name of the column storing a timestamp associated with each row. It is
        used as the default column by the builder when creating watermarks.
    view_definition:
        SQL text defining the view (for `table_type == "VIEW"`). Not supported
        for MANAGED or EXTERNAL table_type.
    warehouse_id:
        All table CRUD operations must be executed on a running cluster or SQL
        warehouse. If a warehouse_id is specified, that SQL warehouse will be
        used to execute SQL commands to manage this table.

    Examples
    --------
    ```py
    from laktory import dlt
    from laktory import models

    dlt.spark = spark

    table = models.Table(
        name="slv_stock_prices",
        builder={
            "layer": "SILVER",
            "table_source": {
                "name": "brz_stock_prices",
            },
            "spark_chain": {
                "nodes": [
                    {
                        "column": {"name": "symbol", "type": "string"},
                        "sql_expression": "data.symbol",
                    },
                    {
                        "column": {
                            "name": "open",
                            "type": "double",
                        },
                        "spark_func_name": "coalesce",
                        "spark_func_args": ["daa.open"],
                    },
                    {
                        "column": {
                            "name": "close",
                            "type": "double",
                        },
                        "spark_func_name": "coalesce",
                        "spark_func_args": ["daa.close"],
                    },
                ]
            },
        },
    )

    # Read
    # df = table.builder.read_source(spark)

    # Process
    # df = table.builder.process(df, None)
    ```

    References
    ----------

    * [Databricks Unity Table](https://docs.databricks.com/en/data-governance/unity-catalog/index.html#tables)
    * [Pulumi Databricks Table](https://www.pulumi.com/registry/packages/databricks/api-docs/sqltable/)
    """

    builder: TableBuilder = TableBuilder()
    catalog_name: Union[str, None] = None
    columns: list[Column] = []
    comment: Union[str, None] = None
    data: list[list[Any]] = None
    data_source_format: str = "DELTA"
    expectations: list[TableExpectation] = []
    grants: list[TableGrant] = None
    name: str
    primary_key: Union[str, None] = None
    schema_name: Union[str, None] = None
    table_type: Literal["MANAGED", "EXTERNA", "VIEW"] = "MANAGED"
    timestamp_key: Union[str, None] = None
    view_definition: str = None
    warehouse_id: str = None

    # ----------------------------------------------------------------------- #
    # Validators                                                              #
    # ----------------------------------------------------------------------- #

    @model_validator(mode="after")
    def assign_catalog_schema(self) -> Any:
        # Assign to columns
        for c in self.columns:
            c.table_name = self.name
            c.catalog_name = self.catalog_name
            c.schema_name = self.schema_name

        # Set builder table
        self.builder._table = self

        # Find table sources
        sources = []
        if self.builder.table_source is not None:
            sources += [self.builder.table_source]

        if self.builder.spark_chain:
            for n in self.builder.spark_chain.nodes:
                for a in n.spark_func_args:
                    if isinstance(a.value, TableDataSource):
                        sources += [a.value]
                for a in n.spark_func_kwargs.values():
                    if isinstance(a.value, TableDataSource):
                        sources += [a.value]

        # Assign catalog and schema to sources
        for s in sources:
            if s.catalog_name is None:
                s.catalog_name = self.catalog_name
            if s.schema_name is None:
                s.schema_name = self.schema_name

        # Warehouse ID
        if self.warehouse_id is None:
            self.warehouse_id = settings.databricks_warehouse_id

        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def parent_full_name(self) -> str:
        """Schema full name `{catalog_name}.{schema_name}`"""
        _id = ""
        if self.catalog_name:
            _id += self.catalog_name

        if self.schema_name:
            if _id == "":
                _id = self.schema_name
            else:
                _id += f".{self.schema_name}"

        return _id

    @property
    def full_name(self) -> str:
        """Table full name `{catalog_name}.{schema_name}.{table_name}`"""
        _id = self.name
        if self.parent_full_name is not None:
            _id = f"{self.parent_full_name}.{_id}"
        return _id

    @property
    def layer(self) -> str:
        """Layer in the medallion architecture ("BRONZE", "SILVER", "GOLD")"""
        return self.builder.layer

    @property
    def column_names(self) -> list[str]:
        """List of column names"""
        return [c.name for c in self.columns]

    @property
    def is_from_cdc(self) -> bool:
        """If `True` CDC source is used to build the table"""
        return self.builder.is_from_cdc

    @property
    def warning_expectations(self) -> dict[str, str]:
        expectations = {}
        for e in self.expectations:
            if e.action == "WARN":
                expectations[e.name] = e.expression
        return expectations

    @property
    def drop_expectations(self) -> dict[str, str]:
        expectations = {}
        for e in self.expectations:
            if e.action == "DROP":
                expectations[e.name] = e.expression
        return expectations

    @property
    def fail_expectations(self) -> dict[str, str]:
        expectations = {}
        for e in self.expectations:
            if e.action == "FAIL":
                expectations[e.name] = e.expression
        return expectations

    # ----------------------------------------------------------------------- #
    #  Methods                                                                #
    # ----------------------------------------------------------------------- #

    def to_df(self, spark=None):
        """
        Dataframe representation of the table. Requires `self.data` to be
        specified.

        Attributes
        ----------
        spark: SparkSession
            Spark session used to convert pandas DataFrame into a spark
            DataFrame if provided.
        """
        import pandas as pd

        df = pd.DataFrame(data=self.data, columns=self.column_names)

        if spark:
            df = spark.createDataFrame(df)
        return df

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        """Table full name (catalog.schema.table)"""
        return self.full_name

    @property
    def self_as_core_resources(self):
        return not self.builder.pipeline_name

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - table grants
        """
        resources = []

        # Schema grants
        if self.grants:
            resources += [
                Grants(
                    resource_name=f"grants-{self.resource_name}",
                    table=self.full_name,
                    grants=[
                        {"principal": g.principal, "privileges": g.privileges}
                        for g in self.grants
                    ],
                )
            ]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:SqlTable"

    @property
    def pulumi_cls(self):
        import pulumi_databricks as databricks

        return databricks.SqlTable

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return [
            "builder",
            "columns",
            "data",
            "expectations",
            "grants",
            "primary_key",
            "timestamp_key",
        ]

    @property
    def pulumi_properties(self):
        d = super().pulumi_properties
        d["columns"] = []
        for i, c in enumerate(self.columns):
            d["columns"] += [
                {
                    "name": c.name,
                    "comment": c.comment,
                    "type": c.type,
                }
            ]
        return d

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_sql_table"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes

    @property
    def terraform_properties(self) -> dict:
        d = super().terraform_properties
        d["column"] = []
        for i, c in enumerate(self.columns):
            d["column"] += [
                {
                    "name": c.name,
                    "comment": c.comment,
                    "type": c.type,
                }
            ]
        return d



from datetime import datetime
from datetime import date
from planck import units
import numpy as np

from laktory import unix_timestamp
from laktory import utc_datetime


def test_unix_timestamp():
    utstamp = 1577840400  # Wednesday, January 1, 2020 1:00:00 AM
    utstamp0 = 1577836800  # Wednesday, January 1, 2020 0:00:00 AM
    dt = datetime(2020, 1, 1, 1, 0, 0)

    assert utc_datetime(utstamp) == dt
    assert unix_timestamp(utc_datetime(utstamp)) == utstamp
    assert unix_timestamp(date(2020, 1, 1)) == utstamp0
    assert utc_datetime(dt) == dt
    assert unix_timestamp(utstamp) == utstamp
    assert utc_datetime("2020-01-01T01:00:00") == dt
    assert unix_timestamp("2020-01-01T01:00:00") == utstamp

    # Null value
    assert utc_datetime(np.datetime64("nat")) is None

    # Timezone
    utstamp0 = unix_timestamp("2022-08-02 15:43:54.199855-02:00")
    utstamp1 = unix_timestamp("2022-08-02T15:43:54.199855+03:00")
    assert utstamp0 - utstamp1 == 5 * units["h"]["s"]


if __name__ == "__main__":
    test_unix_timestamp()



from laktory.models import DbfsFile

dbfs_file = DbfsFile(
    source="./test_workspacefile.py",
    dirpath="/tmp/",
    access_controls=[{"permission_level": "CAN_READ", "group_name": "account users"}],
)


def test_dbfs_file():
    print(dbfs_file)
    assert dbfs_file.filename == "test_workspacefile.py"
    assert dbfs_file.path == "/tmp/test_workspacefile.py"
    assert dbfs_file.resource_key == "tmp-test_workspacefile-py"
    print(dbfs_file.resource_name)
    assert dbfs_file.resource_name == "dbfs-file-tmp-test_workspacefile-py"

    assert dbfs_file.access_controls[0].permission_level == "CAN_READ"
    assert dbfs_file.access_controls[0].group_name == "account users"


if __name__ == "__main__":
    test_dbfs_file()



from laktory.models.basemodel import BaseModel
from laktory.models.providers.baseprovider import BaseProvider
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class ProviderDefaultTags(BaseModel):
    tags: dict[str, str] = None


class ProviderIgnoreTagsArgs(BaseModel):
    key_prefixes: list[str] = None
    keys: list[str] = None


class ProviderAssumeRole(BaseModel):
    duration: str = None
    external_id: str = None
    policy: str = None
    policy_arns: list[str] = None
    role_arn: str = None
    session_name: str = None
    source_identity: str = None
    tags: dict[str, str] = None
    transitive_tag_keys: list[str] = None


class ProviderAssumeRoleWithWebIdentity(BaseModel):
    duration: str = None
    policy: str = None
    policy_arns: list[str] = None
    role_arn: str = None
    session_name: str = None
    web_identity_token: str = None
    web_identity_token_file: str = None


class AWSProvider(BaseProvider, PulumiResource, TerraformResource):
    """
    AWS Provider

    Attributes
    ----------
    access_key:
        The access key for API operations. You can retrieve this from the
        'Security & Credentials' section of the AWS console.
    allowed_account_ids:
        #TODO
    assume_role:
        #TODO
    assume_role_with_web_identity:
        #TODO
    custom_ca_bundle:
        File containing custom root and intermediate certificates. Can also be
        configured using the AWS_CA_BUNDLE environment variable.
        (Setting ca_bundle in the shared config file is not supported.)
    default_tags:
        Configuration block with settings to default resource tags across all
        resources.
    ec2_metadata_service_endpoint:
        Address of the EC2 metadata service endpoint to use. Can also be
        configured using the AWS_EC2_METADATA_SERVICE_ENDPOINT environment
        variable.
    ec2_metadata_service_endpoint_mode:
        Protocol to use with EC2 metadata service endpoint.Valid values are
        IPv4 and IPv6. Can also be configured using the
        AWS_EC2_METADATA_SERVICE_ENDPOINT_MODE environment variable.
    endpoints:
        #TODO
    forbidden_account_ids:
        #TODO
    http_proxy:
        URL of a proxy to use for HTTP requests when accessing the AWS API. Can
        also be set using the HTTP_PROXY or http_proxy environment variables.
    https_proxy:
        URL of a proxy to use for HTTPS requests when accessing the AWS API.
        Can also be set using the HTTPS_PROXY or https_proxy environment
        variables.
    ignore_tags:
        Configuration block with settings to ignore resource tags across all
        resources.
    insecure:
        Explicitly allow the provider to perform "insecure" SSL requests. If
        omitted, default value is false
    max_retries:
        The maximum number of times an AWS API request is being executed. If
        the API request still fails, an error is thrown.
    no_proxy:
        Comma-separated list of hosts that should not use HTTP or HTTPS
        proxies. Can also be set using the NO_PROXY or no_proxy environment
        variables.
    profile:
        The profile for API operations. If not set, the default profile
        created with aws configure will be used.
    region:
        The region where AWS operations will take place. Examples are
        us-east-1, us-west-2, etc. It can also be sourced from the following
        environment variables: AWS_REGION, AWS_DEFAULT_REGION
    retry_mode:
        Specifies how retries are attempted. Valid values are standard and
        adaptive. Can also be configured using the AWS_RETRY_MODE environment
        variable.
    s3_us_east1_regional_endpoint:
        Specifies whether S3 API calls in the us-east-1 region use the legacy
         global endpoint or a regional endpoint. Valid values are legacy or
        regional. Can also be configured using the
        AWS_S3_US_EAST_1_REGIONAL_ENDPOINT environment variable or the
        s3_us_east_1_regional_endpoint shared config file parameter
    s3_use_path_style:
        Set this to true to enable the request to use path-style addressing,
        i.e., https://s3.amazonaws.com/BUCKET/KEY. By default, the S3 client
        will use virtual hosted bucket addressing when possible
        (https://BUCKET.s3.amazonaws.com/KEY). Specific to the Amazon S3
        service.
    secret_key:
        The secret key for API operations. You can retrieve this from the
        'Security & Credentials' section of the AWS console.
    shared_config_files:
        List of paths to shared config files. If not set,
        defaults to [~/.aws/config].
    shared_credentials_files:
        List of paths to shared credentials files. If not set,
        defaults to [~/.aws/credentials].
    skip_credentials_validation:
        Skip the credentials validation via STS API. Used for AWS API
        implementations that do not have STS available/implemented.
    skip_metadata_api_check:
        Skip the AWS Metadata API check. Used for AWS API implementations that
        do not have a metadata api endpoint.
    skip_region_validation:
        Skip static validation of region name. Used by users of alternative
        AWS-like APIs or users w/ access to regions that are not public (yet).
    skip_requesting_account_id:
        Skip requesting the account ID. Used for AWS API implementations that
        do not have IAM/STS API and/or metadata API.
    sts_region:
        The region where AWS STS operations will take place. Examples are
        us-east-1 and us-west-2.
    token:
        Session token. A session token is only required if you are using
        temporary security credentials.
    use_dualstack_endpoint:
        Resolve an endpoint with DualStack capability
    use_fips_endpoint:
        Resolve an endpoint with FIPS capability

    Examples
    --------
    ```py
    from laktory import models

    p = models.AWSProvider(
        access_key="${vars.AWS_ACCESS_KEY}",
    )
    ```
    """

    access_key: str = None
    allowed_account_ids: list[str] = None
    assume_role: ProviderAssumeRole = None
    assume_role_with_web_identity: ProviderAssumeRoleWithWebIdentity = None
    custom_ca_bundle: str = None
    default_tags: ProviderDefaultTags = None
    ec2_metadata_service_endpoint: str = None
    ec2_metadata_service_endpoint_mode: str = None
    # endpoints: SequenceProviderEndpointArgs]] = None
    forbidden_account_ids: list[str] = None
    http_proxy: str = None
    https_proxy: str = None
    ignore_tags: ProviderIgnoreTagsArgs = None
    insecure: bool = None
    max_retries: int = None
    no_proxy: str = None
    profile: str = None
    region: str = None
    retry_mode: str = None
    s3_us_east1_regional_endpoint: str = None
    s3_use_path_style: bool = None
    secret_key: str = None
    shared_config_files: list[str] = None
    shared_credentials_files: list[str] = None
    skip_credentials_validation: bool = None
    skip_metadata_api_check: bool = None
    skip_region_validation: bool = None
    skip_requesting_account_id: bool = None
    sts_region: str = None
    token: str = None
    use_dualstack_endpoint: bool = None
    use_fips_endpoint: bool = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    # @property
    # def resource_key(self) -> str:
    #     return self.display_name

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "pulumi:providers:aws"

    @property
    def pulumi_cls(self):
        import pulumi_aws as aws

        return aws.Provider



from pyspark.sql.dataframe import DataFrame

from laktory.spark.dataframe.schema_flat import schema_flat
from laktory.spark.dataframe.has_column import has_column
from laktory.spark.dataframe.show_string import show_string
from laktory.spark.dataframe.smart_join import smart_join
from laktory.spark.dataframe.watermark import watermark
from laktory.spark.dataframe.groupby_and_agg import groupby_and_agg
from laktory.spark.dataframe.window_filter import window_filter


# DataFrame Extensions
DataFrame.schema_flat = schema_flat
DataFrame.has_column = has_column
DataFrame.show_string = show_string
DataFrame.smart_join = smart_join
DataFrame.watermark = watermark
DataFrame.groupby_and_agg = groupby_and_agg
DataFrame.window_filter = window_filter

# Spark Connect DataFrame Extensions
try:
    from pyspark.sql.connect.dataframe import DataFrame

    DataFrame.schema_flat = schema_flat
    DataFrame.has_column = has_column
    DataFrame.show_string = show_string
    DataFrame.smart_join = smart_join
    DataFrame.watermark = watermark
    DataFrame.groupby_and_agg = groupby_and_agg
    DataFrame.window_filter = window_filter
except:
    pass



from pydantic import model_validator
from typing import Any
from typing import Literal
from typing import Union

from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel
from laktory.models.datasources.basedatasource import BaseDataSource
from laktory.models.datasources import EventDataSource
from laktory.models.datasources import TableDataSource
from laktory.models.sql.column import Column
from laktory.models.spark.sparkchain import SparkChain
from laktory.models.spark.sparkchainnode import SparkChainNode
from laktory.spark import DataFrame

logger = get_logger(__name__)


class TableBuilder(BaseModel):
    """
    Mechanisms for building a table from a source in the context of a data
    pipeline.

    Attributes
    ----------
    add_laktory_columns:
        Add laktory default columns like layer-specific timestamps
    as_dlt_view:
        Create (DLT) view instead of table
    drop_duplicates:
        If `True`:
            - drop duplicated rows using `primary_key` if defined or all
              columns if not defined.
        If list of strings:
            - drop duplicated rows using `drop_duplicates` as the subset.
    drop_source_columns:
        If `True`, drop columns from the source after read and only keep
        columns defined in the table and/or resulting from the joins.
    event_source:
        Definition of the event data source if applicable
    layer:
        Layer in the medallion architecture
    pipeline_name:
        Name of the pipeline in which the table will be built
    spark_chain:
        Spark chain defining the data transformations applied to the data
        source
    table_source:
        Definition of the table data source if applicable
    template:
        Key indicating which notebook to use for building the table in the
        context of a data pipeline.
    """

    add_laktory_columns: bool = True
    as_dlt_view: bool = False
    drop_duplicates: Union[bool, None] = None
    drop_source_columns: Union[bool, None] = None
    event_source: Union[EventDataSource, None] = None
    layer: Literal["BRONZE", "SILVER", "GOLD"] = None
    pipeline_name: Union[str, None] = None
    table_source: Union[TableDataSource, None] = None
    template: Union[str, bool, None] = None
    spark_chain: Union[SparkChain, None] = None
    _table: Any = None

    @model_validator(mode="after")
    def default_values(self) -> Any:
        """
        Sets default options like `drop_source_columns`, `drop_duplicates`,
        `template`, etc. based on `layer` value.
        """
        # Default values
        if self.layer == "BRONZE":
            if self.drop_source_columns is None:
                self.drop_source_columns = False
            if self.drop_duplicates is not None:
                self.drop_duplicates = False

        if self.layer == "SILVER":
            if self.drop_source_columns is None:
                self.drop_source_columns = True
            if self.drop_duplicates is not None and self.primary_key:
                self.drop_duplicates = True
        #
        # if self.layer == "SILVER_STAR":
        #     if self.drop_source_columns is None:
        #         self.drop_source_columns = False
        #     if self.drop_duplicates is not None:
        #         self.drop_duplicates = False
        #
        # if self.layer == "GOLD":
        #     if self.drop_source_columns is None:
        #         self.drop_source_columns = False
        #     if self.drop_duplicates is not None:
        #         self.drop_duplicates = False
        #
        if self.template is None:
            self.template = self.layer

        return self

    @property
    def source(self) -> BaseDataSource:
        """Selected data source"""
        if self.event_source is not None and self.event_source.name is not None:
            return self.event_source
        elif (
            self.table_source is not None
            and self.table_source.path_or_full_name is not None
        ):
            return self.table_source

    @property
    def is_from_cdc(self) -> bool:
        """If `True` CDC source is used to build the table"""
        if self.source is None:
            return False
        else:
            return self.source.is_cdc

    @property
    def columns(self) -> list[Column]:
        """List of columns"""
        if self._table is None:
            return []
        return self._table.columns

    @property
    def timestamp_key(self) -> str:
        """Table timestamp key"""
        if self._table is None:
            return None
        return self._table.timestamp_key

    @property
    def primary_key(self) -> str:
        """Table primary key"""
        if self._table is None:
            return None
        return self._table.primary_key

    @property
    def layer_spark_chain(self):
        nodes = []

        if self.layer == "BRONZE":
            if self.add_laktory_columns:
                nodes += [
                    SparkChainNode(
                        column={
                            "name": "_bronze_at",
                            "type": "timestamp",
                        },
                        spark_func_name="current_timestamp",
                    ),
                ]

        elif self.layer == "SILVER":
            if self.timestamp_key:
                nodes += [
                    SparkChainNode(
                        column={
                            "name": "_tstamp",
                            "type": "timestamp",
                        },
                        sql_expression=self.timestamp_key,
                    )
                ]

            if self.add_laktory_columns:
                nodes += [
                    SparkChainNode(
                        column={
                            "name": "_silver_at",
                            "type": "timestamp",
                        },
                        spark_func_name="current_timestamp",
                    )
                ]

        elif self.layer == "GOLD":
            if self.add_laktory_columns:
                nodes += [
                    SparkChainNode(
                        column={
                            "name": "_gold_at",
                            "type": "timestamp",
                        },
                        spark_func_name="current_timestamp",
                    )
                ]

        if self.drop_duplicates:
            subset = None
            if isinstance(self.drop_duplicates, list):
                subset = self.drop_duplicates
            elif self.primary_key:
                subset = [self.primary_key]

            nodes += [
                SparkChainNode(
                    spark_func_name="dropDuplicates", spark_func_args=[subset]
                )
            ]

        if self.drop_source_columns:
            nodes += [
                SparkChainNode(
                    spark_func_name="drop",
                    spark_func_args=[
                        c
                        for c in self.spark_chain.columns[0]
                        if c not in ["_bronze_at", "_silver_at", "_gold_at"]
                    ],
                )
            ]

        if len(nodes) == 0:
            return None

        return SparkChain(nodes=nodes)

    def read_source(self, spark) -> DataFrame:
        """
        Read data source specified in `self.source`

        Parameters
        ----------
        spark: SparkSession
            Spark session

        Returns
        -------
        :
            Output dataframe
        """
        return self.source.read(spark)

    def process(self, df, udfs=None, spark=None) -> DataFrame:
        """
        Build table by reading source and applying Spark Chain.

        Parameters
        ----------
        df:
            Input DataFrame
        udfs:
            User-defined functions
        spark: SparkSession
            Spark sessions.

        Returns
        -------
        :
            output Spark DataFrame
        """
        logger.info(f"Applying {self.layer} transformations")

        if self.spark_chain is None:
            return df

        df = self.spark_chain.execute(df, udfs=udfs, spark=spark)
        if self.layer_spark_chain:
            df = self.layer_spark_chain.execute(df, udfs=udfs, spark=spark)

        return df

    @property
    def apply_changes_kwargs(self) -> dict[str, str]:
        """Keyword arguments for dlt.apply_changes function"""
        cdc = self.source.cdc
        return {
            "apply_as_deletes": cdc.apply_as_deletes,
            "apply_as_truncates": cdc.apply_as_truncates,
            "column_list": cdc.columns,
            "except_column_list": cdc.except_columns,
            "ignore_null_updates": cdc.ignore_null_updates,
            "keys": cdc.primary_keys,
            "sequence_by": cdc.sequence_by,
            "source": self.source.name,
            "stored_as_scd_type": cdc.scd_type,
            "target": self._table.name,
            "track_history_column_list": cdc.track_history_columns,
            "track_history_except_column_list": cdc.track_history_except_columns,
        }



import os
from pyspark.sql import functions as F

from laktory import models
from laktory._testing import spark
from laktory._testing import Paths

paths = Paths(__file__)

# Data
df_brz = spark.read.parquet(os.path.join(paths.data, "./brz_stock_prices"))
df_slv = spark.read.parquet(os.path.join(paths.data, "./slv_stock_prices"))


def test_read_and_process():
    builder = models.TableBuilder(
        event_source={
            "name": "brz_stock_prices",
            "mock_df": df_brz,
        },
        spark_chain={
            "nodes": [
                {
                    "column": {
                        "name": "created_at",
                        "type": "timestamp",
                    },
                    "sql_expression": "data._created_at",
                },
                {
                    "column": {"name": "symbol", "type": "string"},
                    "sql_expression": "data.symbol",
                },
                {
                    "column": {"name": "open", "type": "double"},
                    "sql_expression": "data.open",
                },
                {
                    "column": {"name": "close", "type": "double"},
                    "sql_expression": "data.close",
                },
                {
                    "spark_func_name": "drop",
                    "spark_func_args": [
                        "description",
                        "data",
                        "producer",
                    ],
                },
            ]
        },
    )

    # Read and process
    df = builder.read_source(spark)
    df = builder.process(df)

    # Test
    assert df.columns == ["name", "created_at", "symbol", "open", "close"]
    assert df.count() == 80


def test_bronze():
    builder = models.TableBuilder(
        layer="BRONZE",
        event_source={
            "name": "brz_stock_prices",
            "mock_df": df_brz,
        },
        spark_chain={
            "nodes": [
                {
                    "column": {"name": "symbol", "type": "string"},
                    "sql_expression": "data.symbol",
                },
            ]
        },
    )

    # Read and process
    df = builder.read_source(spark)
    df = builder.process(df)

    # Test
    assert not builder.drop_source_columns
    assert not builder.drop_duplicates
    assert builder.layer == "BRONZE"
    assert builder.template == "BRONZE"
    assert builder.add_laktory_columns
    assert df.columns == [
        "name",
        "description",
        "producer",
        "data",
        "symbol",
        "_bronze_at",
    ]
    assert df.count() == 80


def test_silver():
    df = df_brz.select(df_brz.columns)
    df = df.withColumn("_bronze_at", F.current_timestamp())
    df = df.union(df)

    builder = models.TableBuilder(
        layer="SILVER",
        table_source={
            "name": "slv_stock_prices",
            "mock_df": df,
        },
        spark_chain={
            "nodes": [
                {
                    "column": {"name": "symbol", "type": "string"},
                    "sql_expression": "data.symbol",
                },
            ]
        },
        drop_duplicates=True,
    )

    # Read and process
    df = builder.read_source(spark)
    df = builder.process(df)

    # Test
    assert builder.drop_source_columns
    assert builder.drop_duplicates
    assert builder.layer == "SILVER"
    assert builder.template == "SILVER"
    assert builder.add_laktory_columns
    assert df.columns == ["_bronze_at", "symbol", "_silver_at"]
    assert df.count() == 80


def test_cdc():
    table = models.Table(
        name="brz_users_type1",
        builder={
            "table_source": {
                "name": "brz_users_cdc",
                "cdc": {
                    "primary_keys": ["userId"],
                    "sequence_by": "sequenceNum",
                    "apply_as_deletes": "operation = 'DELETE'",
                    "scd_type": 1,
                    "except_columns": ["operation", "sequenceNum"],
                },
            },
        },
    )

    assert table.builder.apply_changes_kwargs == {
        "apply_as_deletes": "operation = 'DELETE'",
        "apply_as_truncates": None,
        "column_list": [],
        "except_column_list": ["operation", "sequenceNum"],
        "ignore_null_updates": None,
        "keys": ["userId"],
        "sequence_by": "sequenceNum",
        "source": "brz_users_cdc",
        "stored_as_scd_type": 1,
        "target": "brz_users_type1",
        "track_history_column_list": None,
        "track_history_except_column_list": None,
    }
    assert table.builder.is_from_cdc

    # TODO: Run test with demo data
    # from pyspark.sql import SparkSession

    # spark = SparkSession.builder.appName("UnitTesting").getOrCreate()
    #
    # df_cdc = spark.createDataFrame(pd.DataFrame({
    #     "userId": [124, 123, 125, 126, 123, 125, 125, 123],
    #     "name": ["Raul", "Isabel", "Mercedes", "Lily", None, "Mercedes", "Mercedes", "Isabel"],
    #     "city": ["Oaxaca", "Monterrey", "Tijuana", "Cancun", None, "Guadalajara", "Mexicali", "Chihuahua"],
    #     "operation": ["INSERT", "INSERT", "INSERT", "INSERT", "DELETE", "UPDATE", "UPDATE", "UPDATE"],
    #     "sequenceNum": [1, 1, 2, 2, 6, 6, 5, 5],
    # }))
    # df_cdc.show()


if __name__ == "__main__":
    test_read_and_process()
    test_bronze()
    test_silver()
    test_cdc()



import pytest
from pydantic import ValidationError

from laktory.models import Table
from laktory._testing.stockprices import table_slv
from laktory._testing.stockprices import table_slv_join
from laktory._testing.stockprices import spark


def test_model():
    print(table_slv.model_dump())
    data = table_slv.model_dump()
    assert data == {
        "builder": {
            "add_laktory_columns": True,
            "as_dlt_view": False,
            "drop_duplicates": None,
            "drop_source_columns": True,
            "event_source": None,
            "layer": "SILVER",
            "pipeline_name": None,
            "table_source": {
                "drops": None,
                "filter": None,
                "broadcast": False,
                "read_as_stream": True,
                "renames": None,
                "selects": None,
                "spark_chain": None,
                "watermark": None,
                "catalog_name": "dev",
                "cdc": None,
                "fmt": "DELTA",
                "from_pipeline": True,
                "name": "brz_stock_prices",
                "path": None,
                "schema_name": "markets",
            },
            "template": "SILVER",
            "spark_chain": None,
        },
        "catalog_name": "dev",
        "columns": [
            {
                "catalog_name": "dev",
                "comment": None,
                "name": "created_at",
                "pii": None,
                "raise_missing_arg_exception": True,
                "schema_name": "markets",
                "table_name": "slv_stock_prices",
                "type": "timestamp",
                "unit": None,
            },
            {
                "catalog_name": "dev",
                "comment": None,
                "name": "symbol",
                "pii": None,
                "raise_missing_arg_exception": True,
                "schema_name": "markets",
                "table_name": "slv_stock_prices",
                "type": "string",
                "unit": None,
            },
            {
                "catalog_name": "dev",
                "comment": None,
                "name": "open",
                "pii": None,
                "raise_missing_arg_exception": True,
                "schema_name": "markets",
                "table_name": "slv_stock_prices",
                "type": "double",
                "unit": None,
            },
            {
                "catalog_name": "dev",
                "comment": None,
                "name": "close",
                "pii": None,
                "raise_missing_arg_exception": True,
                "schema_name": "markets",
                "table_name": "slv_stock_prices",
                "type": "double",
                "unit": None,
            },
        ],
        "comment": None,
        "data": [
            ["2023-11-01T00:00:00Z", "AAPL", 1, 2],
            ["2023-11-01T01:00:00Z", "AAPL", 3, 4],
            ["2023-11-01T00:00:00Z", "GOOGL", 3, 4],
            ["2023-11-01T01:00:00Z", "GOOGL", 5, 6],
        ],
        "data_source_format": "DELTA",
        "expectations": [
            {"name": "positive_price", "expression": "open > 0", "action": "FAIL"},
            {
                "name": "recent_price",
                "expression": "created_at > '2023-01-01'",
                "action": "DROP",
            },
        ],
        "grants": None,
        "name": "slv_stock_prices",
        "primary_key": None,
        "schema_name": "markets",
        "table_type": "MANAGED",
        "timestamp_key": None,
        "view_definition": None,
        "warehouse_id": "08b717ce051a0261",
    }

    assert not table_slv.is_from_cdc

    assert table_slv.warning_expectations == {}
    assert table_slv.drop_expectations == {"recent_price": "created_at > '2023-01-01'"}
    assert table_slv.fail_expectations == {"positive_price": "open > 0"}

    # Invalid layer
    with pytest.raises(ValidationError):
        Table(name="googl", layer="ROUGE")


def test_data():
    data = table_slv.to_df().to_dict(orient="records")
    print(data)
    assert data == [
        {"created_at": "2023-11-01T00:00:00Z", "symbol": "AAPL", "open": 1, "close": 2},
        {"created_at": "2023-11-01T01:00:00Z", "symbol": "AAPL", "open": 3, "close": 4},
        {
            "created_at": "2023-11-01T00:00:00Z",
            "symbol": "GOOGL",
            "open": 3,
            "close": 4,
        },
        {
            "created_at": "2023-11-01T01:00:00Z",
            "symbol": "GOOGL",
            "open": 5,
            "close": 6,
        },
    ]


def test_table_builder():
    # Test model
    data = table_slv_join.model_dump()
    print(data)
    assert data == {
        "builder": {
            "add_laktory_columns": True,
            "as_dlt_view": False,
            "drop_duplicates": None,
            "drop_source_columns": False,
            "event_source": None,
            "layer": "SILVER",
            "pipeline_name": None,
            "table_source": {
                "drops": None,
                "filter": "created_at = '2023-09-01T00:00:00Z'",
                "broadcast": False,
                "read_as_stream": True,
                "renames": None,
                "selects": None,
                "spark_chain": None,
                "watermark": None,
                "catalog_name": "dev",
                "cdc": None,
                "fmt": "DELTA",
                "from_pipeline": True,
                "name": "slv_stock_prices",
                "path": None,
                "schema_name": "markets",
            },
            "template": "SILVER",
            "spark_chain": {
                "nodes": [
                    {
                        "allow_missing_column_args": False,
                        "column": None,
                        "spark_func_args": [],
                        "spark_func_kwargs": {
                            "other": {
                                "value": {
                                    "drops": None,
                                    "filter": None,
                                    "broadcast": False,
                                    "read_as_stream": True,
                                    "renames": {"symbol2": "symbol"},
                                    "selects": None,
                                    "spark_chain": None,
                                    "watermark": None,
                                    "catalog_name": "dev",
                                    "cdc": None,
                                    "fmt": "DELTA",
                                    "from_pipeline": True,
                                    "name": "slv_stockmeta",
                                    "path": None,
                                    "schema_name": "markets",
                                }
                            },
                            "on": {"value": ["symbol"]},
                        },
                        "spark_func_name": "smart_join",
                        "sql_expression": None,
                    },
                    {
                        "allow_missing_column_args": False,
                        "column": {"name": "symbol3", "type": "string", "unit": None},
                        "spark_func_args": [],
                        "spark_func_kwargs": {},
                        "spark_func_name": None,
                        "sql_expression": "symbol",
                    },
                    {
                        "allow_missing_column_args": False,
                        "column": None,
                        "spark_func_args": [{"value": "symbol"}],
                        "spark_func_kwargs": {},
                        "spark_func_name": "drop",
                        "sql_expression": None,
                    },
                    {
                        "allow_missing_column_args": False,
                        "column": None,
                        "spark_func_args": [],
                        "spark_func_kwargs": {
                            "other": {
                                "value": {
                                    "drops": None,
                                    "filter": None,
                                    "broadcast": False,
                                    "read_as_stream": True,
                                    "renames": None,
                                    "selects": None,
                                    "spark_chain": None,
                                    "watermark": None,
                                    "catalog_name": "dev",
                                    "cdc": None,
                                    "fmt": "DELTA",
                                    "from_pipeline": True,
                                    "name": "slv_stock_names",
                                    "path": None,
                                    "schema_name": "markets",
                                }
                            },
                            "on": {"value": ["symbol3"]},
                        },
                        "spark_func_name": "smart_join",
                        "sql_expression": None,
                    },
                ]
            },
        },
        "catalog_name": "dev",
        "columns": [],
        "comment": None,
        "data": None,
        "data_source_format": "DELTA",
        "expectations": [],
        "grants": None,
        "name": "slv_join_stock_prices",
        "primary_key": None,
        "schema_name": "markets",
        "table_type": "MANAGED",
        "timestamp_key": None,
        "view_definition": None,
        "warehouse_id": "08b717ce051a0261",
    }

    # Test process
    df = table_slv_join.builder.read_source(spark)
    df = table_slv_join.builder.process(df, spark=spark)
    df = df.sort("symbol3")
    pdf = df.toPandas()

    # Test data
    assert df.columns == [
        "created_at",
        "open",
        "close",
        "currency",
        "first_traded",
        "name",
        "symbol3",
        "_silver_at",
    ]
    assert df.count() == 4
    assert pdf["symbol3"].tolist() == ["AAPL", "AMZN", "GOOGL", "MSFT"]
    assert pdf["name"].tolist() == ["Apple", "Amazon", "Google", None]


if __name__ == "__main__":
    test_model()
    test_data()
    test_table_builder()



import os

from laktory import models
from laktory import Dispatcher
from laktory._version import VERSION
from laktory._testing import Paths

paths = Paths(__file__)

with open(os.path.join(paths.data, "stack.yaml"), "r") as fp:
    stack = models.Stack.model_validate_yaml(fp)


tstack = stack.model_copy()
tstack.backend = "terraform"


def test_workspace_client():
    for _stack in [stack, tstack]:
        dispatcher = Dispatcher(stack=_stack)
        assert f"okube-laktory/{VERSION}" in dispatcher.wc.config.user_agent

        kwargs = dispatcher._workspace_arguments
        if "token" in kwargs:
            kwargs["token"] = kwargs["token"][:6]
        assert kwargs == {
            "host": "https://adb-2211091707396001.1.azuredatabricks.net/",
            "token": "dapic5",
        }
        assert dispatcher.wc is not None


def test_resources():
    dispatcher = Dispatcher(stack=stack)

    assert list(dispatcher.resources.keys()) == [
        "pl-custom-name",
        "job-stock-prices-ut-stack",
    ]
    job = dispatcher.resources["job-stock-prices-ut-stack"]
    pl = dispatcher.resources["pl-custom-name"]

    assert job.model_dump() == {"name": "job-stock-prices-ut-stack", "id": None}
    assert pl.model_dump() == {"name": "pl-stock-prices-ut-stack", "id": None}


if __name__ == "__main__":
    test_workspace_client()
    test_resources()



from laktory.models.resources.baseresource import BaseResource


class BaseResource2(BaseResource):
    pass


def test_resource_name():
    r1 = BaseResource()
    dump = r1.model_dump(exclude_none=True)
    if r1.resource_name_ is not None:
        dump["resource_name_"] = r1.resource_name_
    r1_ = BaseResource2(**dump)

    assert r1.resource_name_ is None
    assert r1.default_resource_name == "base-resource"
    assert r1.resource_name == "base-resource"
    assert r1_.resource_name_ is None
    assert r1_.default_resource_name == "base-resource2"
    assert r1_.resource_name == "base-resource2"

    r2 = BaseResource(resource_name="r2")
    dump = r2.model_dump(exclude_none=True)
    if r2.resource_name_ is not None:
        dump["resource_name_"] = r2.resource_name_
    r2_ = BaseResource2(**dump)
    assert r2.resource_name_ == "r2"
    assert r2.default_resource_name == "base-resource"
    assert r2.resource_name == "r2"
    assert r2_.resource_name_ == "r2"
    assert r2_.default_resource_name == "base-resource2"
    assert r2_.resource_name == "r2"


if __name__ == "__main__":
    test_resource_name()



import typer
import yaml
import os
from typing import Annotated
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit import prompt

from laktory.cli._common import BackendValidator
from laktory.cli._common import DIRPATH
from laktory.cli.app import app
from laktory._parsers import remove_empty
from laktory.models.stacks.stack import Stack
from laktory.constants import SUPPORTED_BACKENDS


def read_template():

    filepath = os.path.join(os.path.dirname(__file__), "quickstart_stack.yaml")

    with open(filepath, "r") as fp:
        stack = Stack.model_validate_yaml(fp)

    return stack


@app.command()
def quickstart(
    backend: Annotated[
        str, typer.Option("--backend", "-b", help="IaC backend [terraform, terraform]")
    ] = None,
    organization: Annotated[
        str,
        typer.Option(
            "--org",
            "-o",
            help="Name of the organization in associated with the pulumi stack.",
        ),
    ] = None,
    node_type: Annotated[
        str,
        typer.Option(
            "--node",
            "-n",
            help="Type of nodes for compute (e.g.: Standard_DS3_v2, c5.2xlarge,)",
        ),
    ] = None,
    filepath: Annotated[
        str, typer.Option(help="Stack (yaml) filepath.")
    ] = "./stack.yaml",
):
    """
    Build get started stack in the calling directory.

    Parameters
    ----------
    backend:
        IaC backend [pulumi, terraform]
    organization:
        Name of the organization associated with the Pulumi stack.
    node_type:
        Type of nodes for compute (e.g.: Standard_DS3_v2, c5.2xlarge,).
    filepath:
        Filepath of the generated stack yaml file.

    Examples
    --------
    ```cmd
    laktory quickstart
    ```
    """

    # Backend
    completer = WordCompleter(SUPPORTED_BACKENDS, ignore_case=True)
    if backend is None:
        backend = prompt(
            f"Select IaC backend {SUPPORTED_BACKENDS}: ",
            completer=completer,
            validator=BackendValidator(),
        )

    # Stack
    stack = "quickstart"

    # Organization
    if backend == "pulumi":
        if organization is None:
            organization = prompt(f"Pulumi Organization: ")
        if stack is None:
            stack = prompt(f"Stack Name (must match Pulumi project name): ")
    elif backend == "terraform":
        if stack is None:
            stack = prompt(f"Stack Name: ")
    else:
        raise ValueError(f"Supported backends are {SUPPORTED_BACKENDS}")

    # Node type id
    if node_type is None:
        node_type = prompt(
            f"Node type for pipeline (e.g.: Standard_DS3_v2, c5.2xlarge, etc.): "
        )

    if not os.path.exists("./notebooks/pipelines"):
        os.makedirs("./notebooks/pipelines")
    if not os.path.exists("./data"):
        os.makedirs("./data")

    # Copy notebooks
    for filename in [
        "dlt_brz_template.py",
        "dlt_slv_template.py",
    ]:
        with open(os.path.join(DIRPATH, f"../resources/notebooks/{filename}")) as fp:
            data = fp.read()
            data = data.replace("pl-stock-prices", "pl-quickstart")

        with open(f"./notebooks/pipelines/{filename}", "w") as fp:
            fp.write(data)

    # Copy data
    for filename in [
        "stock_prices.json",
    ]:
        with open(os.path.join(DIRPATH, f"../resources/data/{filename}")) as fp:
            data = fp.read()

        with open(f"./data/{filename}", "w") as fp:
            fp.write(data)

    # Read template stack
    template_stack = read_template()
    template_stack.backend = backend
    template_stack.resources.pipelines["pl-quickstart"].clusters[
        0
    ].node_type_id = node_type

    # Build stack
    if backend == "pulumi":

        template_stack.resources.providers = {}
        stack = Stack(
            organization=template_stack.organization,
            name=stack,
            backend=backend,
            pulumi={
                "config": {
                    "databricks:host": "${vars.DATABRICKS_HOST}",
                    "databricks:token": "${vars.DATABRICKS_TOKEN}",
                }
            },
            resources=template_stack.resources,
            environments=template_stack.environments,
        )
        stack.pulumi.outputs = None

    else:
        stack = Stack(
            organization=template_stack.organization,
            name=stack,
            backend=backend,
            resources=template_stack.resources,
            environments=template_stack.environments,
        )

    # Dump data
    data = stack.model_dump(exclude_none=True)
    data = remove_empty(data)

    # Clean up resources
    if backend != "pulumi" and "pulumi" in data.keys():
        del data["pulumi"]
    for k, n in data["resources"]["notebooks"].items():
        del data["resources"]["notebooks"][k]["path"]

    # Sort data
    d = {}
    for k in [
        "name",
        "organization",
        "backend",
        "pulumi",
        "resources",
        "environments",
    ]:
        if k in data:
            d[k] = data[k]

    for k in data.keys():
        if k not in d:
            d[k] = data[k]

    # Output Stack
    with open(filepath, "w") as fp:
        yaml.safe_dump(d, fp, sort_keys=False)



import os
import platform
import sys
from laktory._version import VERSION


def show_version_info() -> str:
    from importlib.metadata import distributions

    package_names = {
        "azure-identity",
        "azure-storage-blob",
        "boto3",
        "pandas",
        "planck",
        "pulumi",
        "pulumi_aws",
        "pulumi_azure",
        "pulumi_azure_native",
        "pulumi_databricks",
        "pyarrow",
        "pydantic",
        "pyspark",
        "settus",
    }
    packages = {}

    for d in distributions():
        name = d.metadata["Name"]
        if name in package_names:
            packages[name] = d.version
    packages = dict(sorted(packages.items()))

    _packages = {
        "laktory": VERSION,
        "path": os.path.abspath(os.path.dirname(__file__)),
        "python": sys.version.replace("\n", " - "),
        "platform": platform.platform(),
        "": "-------------------",
    }

    packages = {**_packages, **packages}

    info = "Laktory version info:\n"
    info += "--------------------\n"
    info += "\n".join([f"{k}-{v}" for k, v in packages.items()])

    return info



class MissingColumnError(Exception):
    """"""

    def __init__(self, message="Column is missing", column_name=None):
        if column_name:
            message = f"Column {column_name} is missing"
        super().__init__(message)


class MissingColumnsError(Exception):
    def __init__(self, message="All columns are missing", column_names=None):
        if column_names:
            message = f"Column {column_names} are missing"
        super().__init__(message)



import os

from laktory import Settings


def test_settings():
    settings0 = Settings()
    settings1 = Settings(workspace_landing_root="/mnt/landing/")
    os.environ["LAKTORY_WORKSPACE_LANDING_ROOT"] = "staging"
    settings2 = Settings()
    del os.environ["LAKTORY_WORKSPACE_LANDING_ROOT"]

    assert settings0.workspace_landing_root == "/Volumes/dev/sources/landing/"
    assert settings1.workspace_landing_root == "/mnt/landing/"
    assert settings2.workspace_landing_root == "staging"


if __name__ == "__main__":
    test_settings()



# TODO: Figure out how to import BaseDataSource without creating a circular
# dependency
# from .basedatasource import BaseDataSource
# from .basedatasource import Watermark
from .eventdatasource import EventDataSource
from .tabledatasource import TableDataSource
from .tabledatasource import TableDataSourceCDC



import os
import shutil
import uuid
from laktory._settings import settings
from laktory import models


class StackValidator:
    def __init__(self, resources, providers=None):
        self.resources = resources
        self.providers = providers

    @property
    def stack(self):
        resources = {
            "providers": {
                "databricks": {
                    "host": "${vars.DATABRICKS_HOST}",
                    "token": "${vars.DATABRICKS_TOKEN}",
                }
            }
        }

        if self.providers:
            for k, v in self.providers.items():
                resources["providers"][k] = v

        for k, v in self.resources.items():
            resources[k] = {r.resource_name: r for r in v}

        return models.Stack(
            organization="okube",
            name="unit-testing",
            backend="pulumi",
            pulumi={
                "config": {
                    "databricks:host": "${vars.DATABRICKS_HOST}",
                    "databricks:token": "${vars.DATABRICKS_TOKEN}",
                }
            },
            resources=resources,
            environments={"dev": {}},
        )

    @property
    def pstack(self):
        return self.stack.to_pulumi("dev")

    @property
    def tstack(self):
        return self.stack.to_terraform("dev")

    def validate(self):
        c0 = settings.cli_raise_external_exceptions
        settings.cli_raise_external_exceptions = True
        cwd = os.getcwd()
        dirpath = f"./tmp_deploy_{uuid.uuid4()}"
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)
        os.chdir(dirpath)
        self.validate_pulumi()
        self.validate_terraform()
        os.chdir(cwd)
        shutil.rmtree(dirpath)
        settings.cli_raise_external_exceptions = c0

    def validate_pulumi(self):
        pass
        # self.pstack.preview(stack="okube/dev")
        # TODO: Re-enable

    def validate_terraform(self):
        self.tstack.init()
        self.tstack.plan()



from laktory.models.grants.cataloggrant import CatalogGrant
from laktory.models.grants.connectiongrant import ConnectionGrant
from laktory.models.grants.schemagrant import SchemaGrant
from laktory.models.grants.externallocationgrant import ExternalLocationGrant
from laktory.models.grants.functiongrant import FunctionGrant
from laktory.models.grants.metastoregrant import MetastoreGrant
from laktory.models.grants.registeredmodelgrant import RegisteredModelGrant
from laktory.models.grants.sharegrant import ShareGrant
from laktory.models.grants.storagecredentialgrant import StorageCredentialGrant
from laktory.models.grants.tablegrant import TableGrant
from laktory.models.grants.viewgrant import ViewGrant
from laktory.models.grants.volumegrant import VolumeGrant



import laktory
from laktory.models import BaseModel
from laktory._parsers import remove_empty


class Camel(BaseModel):
    d: dict = {}
    l: list = []


def test_camel_case():
    camel = Camel(
        d={"this_is_a_test": ["value_a", "value_b", {"key_alpha": 0, "keyBeta": 1}]},
        l=["a", "a-b-c", "class_member"],
    )
    d = laktory._parsers.camelize_keys(camel.model_dump())
    print(d)
    assert d == {
        "d": {"thisIsATest": ["value_a", "value_b", {"keyBeta": 1, "keyAlpha": 0}]},
        "l": ["a", "a-b-c", "class_member"],
    }


def test_remove_empty():
    d = {
        "a": "a",
        "b": {},
        "c": [{}, "c", []],
        "d": [],
    }

    assert remove_empty(d) == {"a": "a", "c": ["c"]}


if __name__ == "__main__":
    test_camel_case()
    test_remove_empty()



from typing import Literal
from typing import Any
from typing import Union
from pydantic import Field
from pydantic import model_validator
from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource
from laktory.models.databricks.secret import Secret
from laktory.models.databricks.secretacl import SecretAcl


class SecretScopePermission(BaseModel):
    """
    Secret scope permission

    Attributes
    ----------
    permission:
        Name of the permission to assign
    principal:
        Name of the service principal to assign the permission to
    """

    permission: Literal["READ", "WRITE", "MANAGE"] = None
    principal: str = None


class SecretScopeKeyvaultMetadata(BaseModel):
    """
    Keyvault specifications when used as a secret scope backend

    Attributes
    ----------
    dns_name:

    resource_id:
        Id of the keyvault resource
    """

    dns_name: str = None
    resource_id: str = None


class SecretScope(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks secret scope

    Attributes
    ----------
    backend_type:
        Backend for managing the secrets inside the scope
    keyvault_metadata:
        Keyvault specifications if used as a scope backend
    name:
        Secret scope name
    permissions:
        Permissions given to the secret scope
    secrets:
        List of secret to add to the scope

    Examples
    --------
    ```py
    from laktory import models

    ss = models.SecretScope(
        name="azure",
        secrets=[
            {"key": "keyvault-url", "value": "https://my-secrets.vault.azure.net/"},
            {"key": "client-id", "value": "f461daa2-c281-4166-bc3e-538b90223184"},
        ],
        permissions=[
            {"permission": "READ", "principal": "role-metastore-admins"},
            {"permission": "READ", "principal": "role-workspace-admins"},
        ],
    )
    ```
    """

    backend_type: Literal["DATABRICKS", "AZURE_KEYVAULT"] = "DATABRICKS"
    keyvault_metadata: SecretScopeKeyvaultMetadata = None
    name: str = Field(...)
    permissions: list[SecretScopePermission] = []
    secrets: list[Secret] = []

    @model_validator(mode="after")
    def set_secrets_scope(self) -> Any:
        for s in self.secrets:
            s.scope = self.name

        return self

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - secret values
        - secret scope permissions (ACL)
        """
        resources = []

        for s in self.secrets:
            resources += [
                Secret(
                    key=s.key,
                    value=s.value,
                    scope=f"${{resources.{self.resource_name}.id}}",
                )
            ]

        for p in self.permissions:
            resources += [
                SecretAcl(
                    resource_name=f"secret-scope-acl-{self.name}-{p.principal}",
                    permission=p.permission,
                    principal=p.principal,
                    scope=self.name,
                )
            ]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Methods                                                          #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:SecretScope"

    @property
    def pulumi_cls(self):
        import pulumi_databricks as databricks

        return databricks.SecretScope

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["permissions", "secrets"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_secret_scope"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from typing import Union
from typing import Any
from pydantic import field_validator
from pydantic import ValidationError

from laktory.models.spark.sparkfuncarg import SparkFuncArg


@field_validator("spark_func_args")
def parse_args(cls, args: list[Union[Any, SparkFuncArg]]) -> list[SparkFuncArg]:
    _args = []
    for a in args:
        try:
            a = SparkFuncArg(**a)
        except (ValidationError, TypeError):
            pass

        if isinstance(a, SparkFuncArg):
            pass
        else:
            a = SparkFuncArg(value=a)
        _args += [a]
    return _args


@field_validator("spark_func_kwargs")
def parse_kwargs(
    cls, kwargs: dict[str, Union[str, SparkFuncArg]]
) -> dict[str, SparkFuncArg]:
    _kwargs = {}
    for k, a in kwargs.items():

        try:
            a = SparkFuncArg(**a)
        except (ValidationError, TypeError):
            pass

        if isinstance(a, SparkFuncArg):
            pass
        else:
            a = SparkFuncArg(value=a)

        _kwargs[k] = a
    return _kwargs



import os

from laktory._logger import get_logger


def test_get_logger():
    # Get logger
    logger = get_logger(__name__)

    # Log info
    logger.info("This is an info log")


if __name__ == "__main__":
    test_get_logger()



from typing import Literal

from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel

logger = get_logger(__name__)


class TableExpectation(BaseModel):
    """
    Specifications for aggregation by group or time window.

    Attributes
    ----------
    name:
        Name of the expectation
    expression:
        SQL expression definition the constraint
    action:
        Action to take when expectation is not met.
        `ALLOW`: Invalid records are written to the target; failure is reported
                 as a metric for the dataset.
        `DROP`: Invalid records are dropped before data is written to the
                target; failure is reported as a metrics for the dataset.
        `FAIL`: Invalid records prevent the update from succeeding. Manual
                intervention is required before re-processing.

    Examples
    --------
    ```py
    from laktory import models

    e = models.TableExpectation(
        name="valid timestamp",
        expression="col(“timestamp”) > '2012-01-01'",
        action="DROP",
    )
    print(e)
    '''
    variables={} name='valid timestamp' expression="col(“timestamp”) > '2012-01-01'" action='DROP'
    '''
    ```

    References
    ----------

    * [DLT Table Expectations](https://docs.databricks.com/en/delta-live-tables/expectations.html)
    """

    name: str
    expression: str
    action: Literal["ALLOW", "DROP", "FAIL"] = "ALLOW"



from laktory.models import Metastore


metastore = Metastore(
    name="metastore-lakehouse",
    cloud="azure",
    storage_root="abfss://metastore@o3stglakehousedev.dfs.core.windows.net/",
    region="eastus",
    force_destroy=True,
    workspace_provider="${resources.provider-workspace-neptune}",
    workspace_assignments=[{"workspace_id": 0}],
    grants=[
        {
            "principal": "role-metastore-admins",
            "privileges": [
                "CREATE_CATALOG",
                "CREATE_CONNECTION",
                "CREATE_EXTERNAL_LOCATION",
                "CREATE_STORAGE_CREDENTIAL",
                "MANAGE_ALLOWLIST",
            ],
        }
    ],
    data_accesses=[
        {
            "name": "lakehouse-dev",
            "azure_managed_identity": {"access_connector_id": "test"},
            "force_destroy": True,
            "is_default": False,
            "grants": [
                {
                    "principal": "role-metastore-admins",
                    "privileges": [
                        "CREATE_EXTERNAL_LOCATION",
                    ],
                }
            ],
        }
    ],
)


def test_metastore():
    data = metastore.model_dump()
    print(data)

    assert data == {
        "cloud": "azure",
        "created_at": None,
        "created_by": None,
        "data_accesses": [
            {
                "aws_iam_role": None,
                "azure_managed_identity": {
                    "access_connector_id": "test",
                    "credential_id": None,
                    "managed_identity_id": None,
                },
                "azure_service_principal": None,
                "comment": None,
                "databricks_gcp_service_account": None,
                "force_destroy": True,
                "force_update": None,
                "gcp_service_account_key": None,
                "grants": [
                    {
                        "principal": "role-metastore-admins",
                        "privileges": ["CREATE_EXTERNAL_LOCATION"],
                    }
                ],
                "is_default": False,
                "metastore_id": None,
                "name": "lakehouse-dev",
                "owner": None,
                "read_only": None,
                "skip_validation": None,
            }
        ],
        "default_data_access_config_id": None,
        "delta_sharing_organization_name": None,
        "delta_sharing_recipient_token_lifetime_in_seconds": None,
        "delta_sharing_scope": None,
        "force_destroy": True,
        "global_metastore_id": None,
        "grants": [
            {
                "principal": "role-metastore-admins",
                "privileges": [
                    "CREATE_CATALOG",
                    "CREATE_CONNECTION",
                    "CREATE_EXTERNAL_LOCATION",
                    "CREATE_STORAGE_CREDENTIAL",
                    "MANAGE_ALLOWLIST",
                ],
            }
        ],
        "metastore_id": None,
        "name": "metastore-lakehouse",
        "owner": None,
        "region": "eastus",
        "storage_root": "abfss://metastore@o3stglakehousedev.dfs.core.windows.net/",
        "storage_root_credential_id": None,
        "updated_at": None,
        "updated_by": None,
        "workspace_provider": "${resources.provider-workspace-neptune}",
        "workspace_assignments": [
            {"default_catalog_name": None, "metastore_id": None, "workspace_id": 0}
        ],
    }


if __name__ == "__main__":
    test_metastore()



from pydantic import field_validator
from typing import Any
from typing import Union
from typing import Callable

from laktory._logger import get_logger
from laktory.constants import SUPPORTED_TYPES
from laktory.models.basemodel import BaseModel
from laktory.models.spark.sparkfuncarg import SparkFuncArg
from laktory.spark import DataFrame
from laktory.spark import Column
from laktory.models.spark._common import parse_args
from laktory.models.spark._common import parse_kwargs
from laktory.exceptions import MissingColumnError
from laktory.exceptions import MissingColumnsError

logger = get_logger(__name__)


# --------------------------------------------------------------------------- #
# Main Class                                                                  #
# --------------------------------------------------------------------------- #


class SparkChainNodeColumn(BaseModel):
    """
    Spark chain node column definition

    Attributes
    ----------
    name:
        Column name
    type:
        Column data type
    unit:
        Column units
    """

    name: str
    type: str = "string"
    unit: Union[str, None] = None

    @field_validator("type")
    def check_type(cls, v: str) -> str:
        if "<" in v:
            return v
        else:
            if v not in SUPPORTED_TYPES:
                raise ValueError(
                    f"Type {v} is not supported. Select one of {SUPPORTED_TYPES}"
                )
        return v


class SparkChainNode(BaseModel):
    """
    SparkChain node that output a DataFrame upon execution. As a convenience,
    `column` can be specified to create a new column. In this case, the spark
    function or sql expression is expected to return a column instead of a
    DataFrame. Each node is executed sequentially in the provided order. A node
    may also be another Spark Chain.

    Attributes
    ----------
    allow_missing_column_args:
        If `True`, spark func column arguments are allowed to be missing
        without raising an exception.
    column:
        Column definition. If not `None`, the spark function or sql expression
        is expected to return a column instead of a dataframe.
    spark_func_args:
        List of arguments to be passed to the spark function.
        To support spark functions expecting column argument, col("x"),
        lit("3") and expr("x*2") can be provided.
    spark_func_kwargs:
        List of keyword arguments to be passed to the spark function.
        To support spark functions expecting column argument, col("x"),
        lit("3") and expr("x*2") can be provided.
    spark_func_name:
        Name of the spark function to build the dataframe. If `column` is
        specified, the spark function should return a column instead. Mutually
         exclusive to `sql_expression`.
    sql_expression:
        Expression defining how to build the column. If `column` is
        specified, the sql expression should define a column instead. Mutually
         exclusive to `spark_func_name`

    Examples
    --------
    ```py
    import pandas as pd
    from laktory import models

    df0 = spark.createDataFrame(pd.DataFrame({"x": [1, 2, 2, 3]}))

    node = models.SparkChainNode(
        column={
            "name": "cosx",
            "type": "double",
        },
        spark_func_name="cos",
        spark_func_args=["x"],
    )
    df = node.execute(df0)

    node = models.SparkChainNode(
        column={
            "name": "xy",
            "type": "double",
        },
        spark_func_name="coalesce",
        spark_func_args=["col('x')", "F.col('y')"],
        allow_missing_column_args=True,
    )
    df = node.execute(df)

    print(df.toPandas().to_string())
    '''
       x      cosx   xy
    0  1  0.540302  1.0
    1  2 -0.416147  2.0
    2  2 -0.416147  2.0
    3  3 -0.989992  3.0
    '''

    node = models.SparkChainNode(
        spark_func_name="drop_duplicates",
        spark_func_args=[["x"]],
    )
    df = node.execute(df0, spark=spark)

    print(df.toPandas().to_string())
    '''
       x
    0  1
    1  2
    2  3
    '''
    ```
    """

    allow_missing_column_args: Union[bool, None] = False
    column: Union[SparkChainNodeColumn, None] = None
    spark_func_args: list[Union[Any, SparkFuncArg]] = []
    spark_func_kwargs: dict[str, Union[Any, SparkFuncArg]] = {}
    spark_func_name: Union[str, None] = None
    sql_expression: Union[str, None] = None

    parse_args = parse_args
    parse_kwargs = parse_kwargs

    @property
    def is_column(self):
        return self.column is not None

    @property
    def id(self):
        if self.is_column:
            return self.column.name
        return "df"

    def add_column(self, df, col):
        return df.withColumn(self.column.name, col)

    # ----------------------------------------------------------------------- #
    # Class Methods                                                           #
    # ----------------------------------------------------------------------- #

    def execute(
        self,
        df: DataFrame,
        udfs: list[Callable[[...], Union[Column, DataFrame]]] = None,
        return_col: bool = False,
        spark=None,
    ) -> Union[DataFrame, Column]:
        """
        Build Spark Column

        Parameters
        ----------
        df:
            Input DataFrame
        udfs:
            User-defined functions
        return_col
            If `True` and column specified, function returns `Column` object
            instead of DataFrame.
        spark:
            Spark Session

        Returns
        -------
            Output DataFrame
        """
        import pyspark.sql.functions as F
        from pyspark.sql.dataframe import DataFrame
        from pyspark.sql.connect.dataframe import DataFrame as DataFrameConnect
        from pyspark.sql import Column
        from laktory.spark.dataframe import has_column
        from laktory.spark import functions as LF
        from laktory.spark import DataFrame as LDF

        if udfs is None:
            udfs = []
        udfs = {f.__name__: f for f in udfs}

        # From SQL expression
        if self.sql_expression:
            if self.is_column:
                logger.info(
                    f"{self.column.name}[{self.column.type}] as `{self.sql_expression}`)"
                )
                col = F.expr(self.sql_expression).alias(self.column.name)
                if self.column.type not in ["_any"]:
                    col = col.cast(self.column.type)
                if return_col:
                    return col
                return self.add_column(df, col)
            else:
                raise NotImplementedError("sql_expression not supported yet")
                #     df.createOrReplaceTempView("_df")
                #     df = spark.sql(self.sql_expression)
                #     return df

        # From Spark Function
        func_name = self.spark_func_name
        if self.spark_func_name is None:
            if self.is_column:
                func_name = "coalesce"
            else:
                raise ValueError(
                    "`spark_func_name` must be specified if `sql_expression` is not specified"
                )

        # Get from UDFs
        f = udfs.get(func_name, None)

        # Get from built-in spark functions
        if f is None:
            if self.is_column:
                f = getattr(F, func_name, None)
            else:
                if isinstance(df, DataFrameConnect):
                    f = getattr(DataFrameConnect, func_name, None)
                else:
                    f = getattr(DataFrame, func_name, None)

        # Get from laktory functions
        if f is None:
            if self.is_column:
                f = getattr(LF, func_name, None)
            else:
                f = getattr(LDF, func_name, None)

        if f is None:
            raise ValueError(f"Function {func_name} is not available")

        _args = self.spark_func_args
        _kwargs = self.spark_func_kwargs

        # Build log
        func_log = f"{func_name}("
        for a in _args:
            func_log += f"{a.value},"
        for k, a in _kwargs.items():
            func_log += f"{k}={a.value},"
        if func_log.endswith(","):
            func_log = func_log[:-1]
        func_log += ")"
        if self.is_column:
            logger.info(f"Column {self.id}[{self.column.type}] as {func_log}")
        else:
            logger.info(f"DataFrame {self.id} as {func_log}")

        # Build args
        args = []
        missing_column_names = []
        for i, _arg in enumerate(_args):
            parg = _arg.eval()
            cname = _arg.value

            if df is not None:
                # Check if explicitly defined columns are available
                if isinstance(parg, Column):
                    cname = str(parg).split("'")[1]
                    if not has_column(df, cname):
                        missing_column_names += [cname]
                        if not self.allow_missing_column_args:
                            logger.error(
                                f"Input column {cname} is missing. Abort building {self.id}"
                            )
                            raise MissingColumnError(cname)
                        else:
                            logger.warn(
                                f"Input column {cname} is missing for building {self.id}"
                            )
                            continue

                else:
                    pass
                    # TODO: We could try to check if the function expect columns and if it's the case assume
                    # that a string argument is intended to be a column name, but I feel this will be a rabbit
                    # hole with tons of exceptions.

            args += [parg]

        if len(_args) > 0 and len(args) < 1:
            raise MissingColumnsError(missing_column_names)

        # Build kwargs
        kwargs = {}
        for k, _arg in _kwargs.items():
            kwargs[k] = _arg.eval(spark)

        # Function call
        logger.info(f"{self.id} as {func_name}({args}, {kwargs})")

        if self.is_column:
            col = f(*args, **kwargs)
            if self.column.type not in ["_any"]:
                col = col.cast(self.column.type)
            if return_col:
                return col
            df = self.add_column(df, col)
        else:
            df = f(df, *args, **kwargs)

        return df



from laktory.models.resources.baseresource import BaseResource
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource



import typer
from typing import Annotated

from laktory.cli._common import CLIController
from laktory.cli.app import app
from laktory.constants import SUPPORTED_BACKENDS


@app.command()
def preview(
    backend: Annotated[
        str, typer.Option(help="IaC backend [pulumi, terraform]")
    ] = None,
    organization: Annotated[
        str,
        typer.Option(
            "--org",
            "-o",
            help="Name of the organization in associated with the pulumi stack.",
        ),
    ] = None,
    environment: Annotated[
        str, typer.Option("--env", "-e", help="Name of the environment")
    ] = None,
    filepath: Annotated[
        str, typer.Option(help="Stack (yaml) filepath.")
    ] = "./stack.yaml",
    pulumi_options: Annotated[
        str,
        typer.Option(
            "--pulumi-options", help="Comma separated pulumi options (flags)."
        ),
    ] = None,
    terraform_options: Annotated[
        str,
        typer.Option(
            "--terraform-options", help="Comma separated terraform options (flags)."
        ),
    ] = None,
):
    """
    Validate configuration and resources and preview deployment.

    Parameters
    ----------
    backend:
        IaC backend [pulumi, terraform]
    organization:
        Name of the organization associated with the Pulumi stack.
    environment:
        Name of the environment.
    filepath:
        Stack (yaml) filepath.
    pulumi_options:
        Comma separated pulumi options (flags).
    terraform_options:
        Comma separated terraform options (flags).

    Examples
    --------
    ```cmd
    laktory preview --env dev pulumi_options "--show-reads,--show-config"
    ```

    References
    ----------
    - pulumi preview [options](https://www.pulumi.com/docs/cli/commands/pulumi_preview/)
    """
    controller = CLIController(
        backend=backend,
        organization=organization,
        env=environment,
        stack_filepath=filepath,
        pulumi_options_str=pulumi_options,
        terraform_options_str=terraform_options,
    )

    # Call
    if controller.backend == "pulumi":
        controller.pulumi_call("preview")
    elif controller.backend == "terraform":
        controller.terraform_call("plan")
    else:
        raise ValueError(f"backend should be {SUPPORTED_BACKENDS}")



from typing import Union

from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel
from laktory.models.spark.sparkchainnode import SparkChainNode
from laktory.spark import DataFrame

logger = get_logger(__name__)


# --------------------------------------------------------------------------- #
# Main Class                                                                  #
# --------------------------------------------------------------------------- #


class SparkChain(BaseModel):
    """
    The Spark Chain class defines a series of transformation to be applied to
    a DataFrame. Each transformation is expressed as a node (SparkChainNode
    object) that, upon execution, returns a new dataframe. As a convenience,
    `column` can be specified to create a new column. In this case, the spark
    function or sql expression is expected to return a column instead of a
    DataFrame. Each node is executed sequentially in the provided order. A node
    may also be another Spark Chain.

    Attributes
    ----------
    nodes:
        The list of transformations to be executed.

    Examples
    --------
    ```py
    import pandas as pd
    from laktory import models

    df0 = spark.createDataFrame(pd.DataFrame({"x": [1, 2, 3]}))

    # Build Chain
    sc = models.SparkChain(
        nodes=[
            {
                "column": {
                    "name": "cos_x",
                    "type": "double",
                },
                "spark_func_name": "cos",
                "spark_func_args": ["x"],
            },
            {
                "nodes": [
                    {
                        "spark_func_name": "withColumnRenamed",
                        "spark_func_args": [
                            "x",
                            "x_tmp",
                        ],
                    },
                    {
                        "column": {
                            "name": "x2",
                            "type": "double",
                        },
                        "spark_func_name": "sqrt",
                        "spark_func_args": ["x_tmp"],
                    },
                ],
            },
            {
                "spark_func_name": "drop",
                "spark_func_args": [
                    "x_tmp",
                ],
            },
        ]
    )

    # Execute Chain
    df = sc.execute(df0, spark=spark)

    # Print result
    print(df.toPandas().to_string())
    '''
          cos_x        x2
    0  0.540302  1.000000
    1 -0.416147  1.414214
    2 -0.989992  1.732051
    '''
    ```
    """

    nodes: list[Union[SparkChainNode, "SparkChain"]]
    _columns: list[list[str]] = []

    @property
    def columns(self):
        return self._columns

    def execute(self, df, udfs=None, spark=None) -> DataFrame:
        logger.info("Executing Spark chain")

        for inode, node in enumerate(self.nodes):
            self._columns += [df.columns]

            tnode = type(node)
            logger.info(f"Executing node {inode} ({tnode.__name__}).")
            df = node.execute(df, udfs=udfs, spark=spark)

        return df


SparkChain.model_rebuild()


def _builtin_dataframe_functions():
    from pyspark.sql import DataFrame
    import inspect

    func_names = []

    for k in vars(DataFrame).keys():
        if k.startswith("_"):
            continue
        m = getattr(DataFrame, k)

        if not (inspect.isfunction(m) or inspect.ismethod(m)):
            continue

        sig = inspect.signature(m)
        return_type = sig.return_annotation

        if return_type == "DataFrame":
            func_names += [k]

    return func_names



import pandas as pd
import numpy as np
import pytest
from pyspark.sql import SparkSession
from pyspark.sql import functions as F
from pyspark.sql import types as T

from laktory import models
from laktory.exceptions import MissingColumnError
from laktory.exceptions import MissingColumnsError

pdf = pd.DataFrame(
    {
        "x": [1, 2, 3],
        "a": [1, -1, 1],
        "b": [2, 0, 2],
        "c": [3, 0, 3],
        "n": [4, 0, 4],
        "pi": [np.pi] * 3,
        "p": [2, 0.2, 0.05],
        "word": ["dog_cat", "dog_cat_mouse", "dog"],
    },
)
spark = SparkSession.builder.appName("UnitTesting").getOrCreate()

df0 = spark.createDataFrame(pdf)
df1 = spark.createDataFrame(pdf)


def test_spark_func_arg(df0=df0):
    df = df0.select(df0.columns)

    new_cols = []
    # Values to parse
    for i, v in enumerate(
        [
            "lit(3)",
            "lit(3.0)",
            "lit('3')",
            "col('x')",
            "col('x') + lit(3)",
            "expr('2*x+a')",
        ]
    ):
        new_col = f"c{i}"
        a = models.SparkFuncArg(value=v)
        df = df.withColumn(new_col, a.eval(spark))
        new_cols += [new_col]

    df.show()

    # Test new column types
    assert df.select(new_cols).schema == T.StructType(
        [
            T.StructField("c0", T.IntegerType(), False),
            T.StructField("c1", T.DoubleType(), False),
            T.StructField("c2", T.StringType(), False),
            T.StructField("c3", T.LongType(), True),
            T.StructField("c4", T.LongType(), True),
            T.StructField("c5", T.LongType(), True),
        ]
    )

    # Values not to parse
    for v0 in [
        "x",
        "3",
    ]:
        v1 = models.SparkFuncArg(value=v0).eval(spark)
        assert v1 == v0


def test_dataframe_df_input(df0=df0):
    df = df0.select(df0.columns)

    # Define Chain
    sc = models.SparkChain(
        nodes=[
            {
                "spark_func_name": "union",
                "spark_func_args": [df1],
            },
        ]
    )

    # Execute Chain
    df = sc.execute(df)

    # Test
    assert df.count() == df0.count() * 2
    assert df.columns == df0.columns


def test_dataframe_table_input(df0=df0):
    df = df0.select(df0.columns)

    sc = models.SparkChain(
        nodes=[
            {
                "spark_func_name": "union",
                "spark_func_args": [
                    {"mock_df": df},
                ],
            }
        ]
    )

    # Execute Chain
    df = sc.execute(df)

    # Test
    assert df.count() == df0.count() * 2
    assert df.columns == df0.columns


def test_column(df0=df0):
    df = df0.select(df0.columns)

    sc = models.SparkChain(
        nodes=[
            {
                "column": {"name": "cos_x", "type": "double"},
                "spark_func_name": "cos",
                "spark_func_args": ["x"],
            },
            {
                "column": {"name": "x2", "type": "double"},
                "sql_expression": "x*2",
            },
        ]
    )

    # Execute Chain
    df = sc.execute(df)

    # Test
    pdf = df.toPandas()
    assert pdf["cos_x"].tolist() == np.cos(pdf["x"]).tolist()
    assert pdf["x2"].tolist() == (pdf["x"] * 2).tolist()


def test_udfs(df0=df0):
    df = df0.select(df0.columns)

    def mul3(c):
        return 3 * c

    def add_new_col(df, column_name, s=1):
        return df.withColumn(column_name, F.col("x") * s)

    sc = models.SparkChain(
        nodes=[
            {
                "column": {
                    "name": "rp",
                    "type": "double",
                },
                "spark_func_name": "roundp",
                "spark_func_args": ["p"],
                "spark_func_kwargs": {"p": 0.1},
            },
            {
                "column": {"name": "x3", "type": "double"},
                "spark_func_name": "mul3",
                "spark_func_args": [F.col("x")],
            },
            {
                "spark_func_name": "add_new_col",
                "spark_func_args": ["y"],
                "spark_func_kwargs": {"s": 5},
            },
        ]
    )

    # Execute Chain
    df = sc.execute(df, udfs=[mul3, add_new_col])

    # Test
    pdf = df.toPandas()
    assert pdf["rp"].tolist() == [2.0, 0.2, 0.1]
    assert pdf["x3"].tolist() == (pdf["x"] * 3).tolist()
    assert pdf["y"].tolist() == (pdf["x"] * 5).tolist()


def test_nested(df0=df0):
    df = df0.select(df0.columns)
    df = df.withColumn("_x2", F.sqrt("x"))

    sc = models.SparkChain(
        nodes=[
            {
                "column": {"name": "cos_x", "type": "double"},
                "spark_func_name": "cos",
                "spark_func_args": ["x"],
            },
            {
                "nodes": [
                    {
                        "spark_func_name": "withColumnRenamed",
                        "spark_func_args": [
                            "x",
                            "x_tmp",
                        ],
                    },
                    {
                        "column": {"name": "x2", "type": "double"},
                        "spark_func_name": "sqrt",
                        "spark_func_args": ["x_tmp"],
                    },
                ],
            },
            {
                "spark_func_name": "drop",
                "spark_func_args": [
                    "x_tmp",
                ],
            },
        ]
    )

    # Execute Chain
    df = sc.execute(df)

    # Test
    pdf = df.toPandas()
    assert "x_tmp" not in pdf.columns
    assert pdf["x2"].tolist() == pdf["_x2"].tolist()
    assert sc.columns == [
        ["x", "a", "b", "c", "n", "pi", "p", "word", "_x2"],
        ["x", "a", "b", "c", "n", "pi", "p", "word", "_x2", "cos_x"],
        ["x_tmp", "a", "b", "c", "n", "pi", "p", "word", "_x2", "cos_x", "x2"],
    ]


def test_exceptions():
    df = df0.select(df0.columns)

    # Input missing - missing not allowed
    sc = models.SparkChain(
        nodes=[
            {
                "column": {"name": "cos_x", "type": "double"},
                "spark_func_name": "cos",
                "spark_func_args": ["col('y')"],
                "allow_missing_column_args": False,
            },
        ]
    )
    with pytest.raises(MissingColumnError):
        df = sc.execute(df)

    # Input missing - missing allowed
    sc = models.SparkChain(
        nodes=[
            {
                "column": {"name": "xy", "type": "double"},
                "spark_func_name": "coalesce",
                "spark_func_args": ["col('x')", "col('y')"],
                "allow_missing_column_args": True,
            },
        ]
    )
    df = sc.execute(df)
    assert "xy" in df.columns

    # All inputs missing
    sc = models.SparkChain(
        nodes=[
            {
                "column": {"name": "xy", "type": "double"},
                "spark_func_name": "coalesce",
                "spark_func_args": ["col('z')", "col('y')"],
                "allow_missing_column_args": True,
            },
        ]
    )
    with pytest.raises(MissingColumnsError):
        df = sc.execute(df)


if __name__ == "__main__":
    test_spark_func_arg()
    test_dataframe_df_input()
    test_dataframe_table_input()
    test_column()
    test_udfs()
    test_nested()
    test_exceptions()



import os
import argparse


def main(branch_name: str, notebooks_dir: str):

    for filename in os.listdir(notebooks_dir):
        filepath = os.path.join(notebooks_dir, filename)

        print(f"Updating {filepath} with laktory from branch {branch_name}")

        with open(filepath, "r") as fp:
            data = fp.read()

        if "pip install laktory" not in data:
            continue

        data = data.replace(
            "pip install laktory",
            f"pip install git+https://github.com/okube-ai/laktory.git@{branch_name}",
        )

        with open(filepath, "w") as fp:
            fp.write(data)


if __name__ == "__main__":
    # Parse arguments
    parser = argparse.ArgumentParser(
        description="Update notebooks to install specific version of laktory"
    )
    parser.add_argument("branch_name", type=str, help="Laktory branch name")
    parser.add_argument(
        "--notebooks_path",
        type=str,
        help="Path of the notebooks directory",
        default="./notebooks/pipelines",
    )
    args = parser.parse_args()

    # Execute
    main(args.branch_name, args.notebooks_path)



from .catalog import Catalog
from .column import Column
from .schema import Schema
from .table import Table
from .tablebuilder import TableBuilder
from .tableexpectation import TableExpectation
from .volume import Volume



from laktory.models import SqlQuery

query = SqlQuery(
    name="google-prices",
    parent="/queries",
    query="SELECT * FROM dev.finance.slv_stock_prices",
    data_source_id="12345",
)


def test_sql_query():
    data = query.model_dump()
    print(query.resource_key)
    assert query.resource_key == "google-prices"
    print(query.resource_name)
    assert query.resource_name == "sql-query-google-prices"
    print(data)
    assert data == {
        "access_controls": [],
        "comment": None,
        "data_source_id": "12345",
        "name": "google-prices",
        "parent": "/queries",
        "query": "SELECT * FROM dev.finance.slv_stock_prices",
        "run_as_role": None,
        "tags": [],
    }


if __name__ == "__main__":
    test_sql_query()



import typer
from typing import Annotated

from laktory.cli._common import CLIController
from laktory.cli.app import app
from laktory.dispatcher.dispatcher import Dispatcher


@app.command()
def run(
    job: Annotated[str, typer.Option("--job", "-j", help="Job name")] = None,
    pipeline: Annotated[str, typer.Option("--pipeline", "-p", help="Job name")] = None,
    timeout: Annotated[
        float,
        typer.Option(
            "--timeout", "-t", help="Maximum allowed time (in seconds) for run"
        ),
    ] = 1200,
    raise_exception: Annotated[
        bool, typer.Option("--raise", "-r", help="Raise exception on failure")
    ] = True,
    full_refresh: Annotated[
        bool,
        typer.Option(
            "--full-refresh", "--fr", help="Full tables refresh (pipeline only)"
        ),
    ] = False,
    current_run_action: Annotated[
        str,
        typer.Option(
            "--action",
            "-a",
            help="Action to take if job currently running ['WAIT', 'CANCEL', 'FAIL']",
        ),
    ] = "WAIT",
    environment: Annotated[
        str, typer.Option("--env", "-e", help="Name of the environment")
    ] = None,
    filepath: Annotated[
        str, typer.Option(help="Stack (yaml) filepath.")
    ] = "./stack.yaml",
):
    """
    Execute remote job or pipeline and monitor failures until completion.

    Parameters
    ----------
    job:
        Name of the job to run (mutually exclusive with pipeline)
    pipeline:
        Name of the pipeline to run (mutually exclusive with pipeline)
    timeout:
        Maximum allowed time (in seconds) for run.
    raise_exception:
        Raise exception on failure
    current_run_action:
        Action to take for currently running job or pipline.
    full_refresh:
        Full tables refresh (pipline only)
    environment:
        Name of the environment.
    filepath:
        Stack (yaml) filepath.

    Examples
    --------
    ```cmd
    laktory run --env dev -p pl-stock-prices --full_refresh --action CANCEL
    ```

    """

    # Set Resource Name
    if job and pipeline:
        raise ValueError("Only one of `job` or `pipeline` should be set.")
    if not (job or pipeline):
        raise ValueError("One of `job` or `pipeline` should be set.")

    # Set Dispatcher
    controller = CLIController(
        env=environment,
        stack_filepath=filepath,
    )
    dispatcher = Dispatcher(stack=controller.stack)
    dispatcher.get_resource_ids()

    if job:
        dispatcher.run_job(
            job_name=job,
            timeout=timeout,
            raise_exception=raise_exception,
            current_run_action=current_run_action,
        )

    if pipeline:
        dispatcher.run_pipeline(
            pipeline_name=pipeline,
            timeout=timeout,
            raise_exception=raise_exception,
            current_run_action=current_run_action,
            full_refresh=full_refresh,
        )



from abc import abstractmethod
from typing import Union
from laktory._settings import settings
from laktory._parsers import _snake_to_camel
from laktory.models.resources.baseresource import BaseResource


class TerraformResource(BaseResource):
    """
    Parent class for all Laktory models deployable with Terraform IaC backend.
    """

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    @abstractmethod
    def terraform_resource_type(self) -> str:
        raise NotImplementedError()

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        """List of fields to exclude when dumping model to terraform"""
        return []

    @property
    def terraform_renames(self) -> dict[str, str]:
        """Map of fields to rename when dumping model to terraform"""
        return {}

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_properties(self) -> dict:
        """
        Resources properties formatted for terraform:

        * Serialization (model dump)
        * Removal of excludes defined in `self.terraform_excludes`
        * Renaming of keys according to `self.terraform_renames`
        * Injection of variables

        Returns
        -------
        :
            Terraform-safe model dump
        """
        d = super().model_dump(exclude=self.terraform_excludes, exclude_none=True)
        for k, v in self.terraform_renames.items():
            if k in d:
                d[v] = d.pop(k)

        # Add options
        for k in ["depends_on", "provider"]:
            value = self.options.model_dump(exclude_none=True).get(k, None)
            if value:
                d[k] = value

        d = self.inject_vars(d)
        return d



from laktory.models.basemodel import BaseModel
from laktory.models.databricks.accesscontrol import AccessControl
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class Permissions(BaseModel, PulumiResource, TerraformResource):
    access_controls: list[AccessControl]
    pipeline_id: str = None
    job_id: str = None
    cluster_id: str = None
    directory_id: str = None
    directory_path: str = None
    experiment_id: str = None
    notebook_id: str = None
    notebook_path: str = None
    object_type: str = None
    registered_model_id: str = None
    repo_id: str = None
    repo_path: str = None
    serving_endpoint_id: str = None
    sql_alert_id: str = None
    sql_dashboard_id: str = None
    sql_endpoint_id: str = None
    sql_query_id: str = None
    workspace_file_id: str = None
    workspace_file_path: str = None

    # ----------------------------------------------------------------------- #
    # Pulumi Methods                                                          #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Permissions"

    @property
    def pulumi_cls(self):
        import pulumi_databricks as databricks

        return databricks.Permissions

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_permissions"



import os
from laktory import app
from laktory import settings
from laktory import models
from laktory._testing import Paths
from typer.testing import CliRunner

runner = CliRunner()
settings.cli_raise_external_exceptions = True
paths = Paths(__file__)


def test_preview_pulumi():
    filepath = os.path.join(paths.data, "stack.yaml")
    result = runner.invoke(app, ["preview", "--env", "dev", "--filepath", filepath])
    assert result.exit_code == 0


def test_preview_terraform():
    filepath = os.path.join(paths.data, "stack.yaml")

    # Ideally, we would run `laktory init`, but the runner does not seem to handle running multiple commands
    with open(filepath, "r") as fp:
        pstack = models.Stack.model_validate_yaml(fp).to_terraform(env="dev")
        pstack.init(flags=["-migrate-state", "-upgrade"])

    result = runner.invoke(
        app,
        ["preview", "--backend", "terraform", "--env", "dev", "--filepath", filepath],
    )
    assert result.exit_code == 0


def test_quickstart_pulumi():
    filepath = os.path.join(paths.tmp, "stack_quickstart_pulumi.yaml")
    print(filepath)
    result = runner.invoke(
        app,
        [
            "quickstart",
            "--backend",
            "pulumi",
            "-o",
            "okube",
            "-n",
            "Standard_DS3_v2",
            "--filepath",
            filepath,
        ],
    )

    # TODO: Add validation for the generated stack?
    with open(filepath) as fp:
        stack = models.Stack.model_validate_yaml(fp)

    data = stack.model_dump(exclude_none=True)

    assert stack.backend == "pulumi"
    assert stack.name == "quickstart"
    assert stack.pulumi.config["databricks:host"] == "${vars.DATABRICKS_HOST}"
    assert stack.pulumi.config["databricks:token"] == "${vars.DATABRICKS_TOKEN}"
    assert len(stack.resources.dbfsfiles) == 1
    assert len(stack.resources.notebooks) == 2
    assert len(stack.resources.pipelines) == 1


def test_quickstart_terraform():
    filepath = os.path.join(paths.tmp, "stack_quickstart_terraform.yaml")
    result = runner.invoke(
        app,
        [
            "quickstart",
            "--backend",
            "terraform",
            "-o",
            "okube",
            "-n",
            "Standard_DS3_v2",
            "--filepath",
            filepath,
        ],
    )

    with open(filepath) as fp:
        stack = models.Stack.model_validate_yaml(fp)

    data = stack.model_dump(exclude_none=True)

    assert stack.backend == "terraform"
    assert stack.name == "quickstart"
    assert stack.resources.providers["databricks"].host == "${vars.DATABRICKS_HOST}"
    assert stack.resources.providers["databricks"].token == "${vars.DATABRICKS_TOKEN}"
    assert len(stack.resources.dbfsfiles) == 1
    assert len(stack.resources.notebooks) == 2
    assert len(stack.resources.pipelines) == 1


def atest_deploy_pulumi():
    # TODO: Figure out how to run in isolation. Currently, pulumi up commands
    # are run concurrently because of the multiple python testing environment
    # which result in:  Conflict: Another update is currently in progress
    for filename in [
        "stack.yaml",
        "stack_empty.yaml",
    ]:
        filepath = os.path.join(paths.data, filename)
        result = runner.invoke(
            app,
            [
                "deploy",
                "-e",
                "dev",
                "--filepath",
                filepath,
                "--pulumi-options",
                "--yes",
            ],
        )
        assert result.exit_code == 0


def atest_deploy_terraform():
    # TODO: Figure out how to run in isolation. Currently, pulumi up commands
    # are run concurrently because of the multiple python testing environment
    # which result in:  Conflict: Another update is currently in progress
    for filename in [
        "stack.yaml",
        "stack_empty.yaml",
    ]:
        filepath = os.path.join(paths.data, filename)
        result = runner.invoke(
            app,
            [
                "deploy",
                "-e",
                "dev",
                "--backend",
                "terraform",
                "--filepath",
                filepath,
                "--terraform-options",
                "--auto-approve",
            ],
        )
        assert result.exit_code == 0


if __name__ == "__main__":
    test_preview_pulumi()
    test_preview_terraform()
    test_quickstart_pulumi()
    test_quickstart_terraform()
    # atest_deploy_pulumi()
    # atest_deploy_terraform()



from pydantic import field_validator
from typing import Any
from typing import Union

from laktory.models.basemodel import BaseModel


# --------------------------------------------------------------------------- #
# Main Class                                                                  #
# --------------------------------------------------------------------------- #


class SparkFuncArg(BaseModel):
    """
    Spark function argument

    Attributes
    ----------
    value:
        Value of the argument
    """

    value: Union[Any]

    @field_validator("value")
    def value_to_data_source(cls, v: Any) -> Any:
        """
        Data source can't be set as an expected value type as it would create
        a circular dependency. Instead, we check at validation if the value
        could be instantiated as a DataSource object.
        """
        from laktory.models.datasources.eventdatasource import EventDataSource
        from laktory.models.datasources.tabledatasource import TableDataSource

        try:
            v = TableDataSource(**v)
        except:
            try:
                v = EventDataSource(**v)
            except:
                pass

        return v

    def eval(self, spark=None):
        from laktory.models.datasources.basedatasource import BaseDataSource

        # Imports required to evaluate expressions
        import pyspark.sql.functions as F
        from pyspark.sql.functions import lit
        from pyspark.sql.functions import col
        from pyspark.sql.functions import expr

        v = self.value

        if isinstance(v, BaseDataSource):
            v = self.value.read(spark)
        elif isinstance(v, str):
            for f in ["lit", "col", "expr", "F."]:
                if f in v:
                    v = eval(v)
                    break

        return v



from laktory import models

from laktory._testing.stockprices import table_slv_pl
from laktory._testing.stockprices import table_slv_join_pl


pl = models.Pipeline(
    name="pl-stock-prices",
    catalog="dev1",
    target="markets1",
    tables=[table_slv_pl, table_slv_join_pl],
    udfs=[
        {
            "module_name": "stock_functions",
            "function_name": "high",
        }
    ],
)


def test_pipeline():
    print(pl.model_dump())
    assert pl.model_dump() == {
        "access_controls": [],
        "allow_duplicate_names": None,
        "catalog": "dev1",
        "channel": "PREVIEW",
        "clusters": [],
        "configuration": {},
        "continuous": None,
        "development": None,
        "edition": None,
        "libraries": None,
        "name": "pl-stock-prices",
        "notifications": [],
        "photon": None,
        "serverless": None,
        "storage": None,
        "tables": [
            {
                "builder": {
                    "add_laktory_columns": True,
                    "as_dlt_view": False,
                    "drop_duplicates": None,
                    "drop_source_columns": True,
                    "event_source": None,
                    "layer": "SILVER",
                    "pipeline_name": "pl-stock-prices",
                    "table_source": {
                        "drops": None,
                        "filter": None,
                        "broadcast": False,
                        "read_as_stream": True,
                        "renames": None,
                        "selects": None,
                        "spark_chain": None,
                        "watermark": None,
                        "catalog_name": "dev",
                        "cdc": None,
                        "fmt": "DELTA",
                        "from_pipeline": True,
                        "name": "brz_stock_prices",
                        "path": None,
                        "schema_name": "markets",
                    },
                    "template": "SILVER",
                    "spark_chain": None,
                },
                "catalog_name": "dev1",
                "columns": [
                    {
                        "catalog_name": "dev1",
                        "comment": None,
                        "name": "created_at",
                        "pii": None,
                        "raise_missing_arg_exception": True,
                        "schema_name": "markets1",
                        "table_name": "slv_stock_prices",
                        "type": "timestamp",
                        "unit": None,
                    },
                    {
                        "catalog_name": "dev1",
                        "comment": None,
                        "name": "symbol",
                        "pii": None,
                        "raise_missing_arg_exception": True,
                        "schema_name": "markets1",
                        "table_name": "slv_stock_prices",
                        "type": "string",
                        "unit": None,
                    },
                    {
                        "catalog_name": "dev1",
                        "comment": None,
                        "name": "open",
                        "pii": None,
                        "raise_missing_arg_exception": True,
                        "schema_name": "markets1",
                        "table_name": "slv_stock_prices",
                        "type": "double",
                        "unit": None,
                    },
                    {
                        "catalog_name": "dev1",
                        "comment": None,
                        "name": "close",
                        "pii": None,
                        "raise_missing_arg_exception": True,
                        "schema_name": "markets1",
                        "table_name": "slv_stock_prices",
                        "type": "double",
                        "unit": None,
                    },
                ],
                "comment": None,
                "data": [
                    ["2023-11-01T00:00:00Z", "AAPL", 1, 2],
                    ["2023-11-01T01:00:00Z", "AAPL", 3, 4],
                    ["2023-11-01T00:00:00Z", "GOOGL", 3, 4],
                    ["2023-11-01T01:00:00Z", "GOOGL", 5, 6],
                ],
                "data_source_format": "DELTA",
                "expectations": [
                    {
                        "name": "positive_price",
                        "expression": "open > 0",
                        "action": "FAIL",
                    },
                    {
                        "name": "recent_price",
                        "expression": "created_at > '2023-01-01'",
                        "action": "DROP",
                    },
                ],
                "grants": None,
                "name": "slv_stock_prices",
                "primary_key": None,
                "schema_name": "markets1",
                "table_type": "MANAGED",
                "timestamp_key": None,
                "view_definition": None,
                "warehouse_id": "08b717ce051a0261",
            },
            {
                "builder": {
                    "add_laktory_columns": True,
                    "as_dlt_view": False,
                    "drop_duplicates": None,
                    "drop_source_columns": False,
                    "event_source": None,
                    "layer": "SILVER",
                    "pipeline_name": "pl-stock-prices",
                    "table_source": {
                        "drops": None,
                        "filter": "created_at = '2023-09-01T00:00:00Z'",
                        "broadcast": False,
                        "read_as_stream": True,
                        "renames": None,
                        "selects": None,
                        "spark_chain": None,
                        "watermark": None,
                        "catalog_name": "dev",
                        "cdc": None,
                        "fmt": "DELTA",
                        "from_pipeline": True,
                        "name": "slv_stock_prices",
                        "path": None,
                        "schema_name": "markets",
                    },
                    "template": "SILVER",
                    "spark_chain": {
                        "nodes": [
                            {
                                "allow_missing_column_args": False,
                                "column": None,
                                "spark_func_args": [],
                                "spark_func_kwargs": {
                                    "other": {
                                        "value": {
                                            "drops": None,
                                            "filter": None,
                                            "broadcast": False,
                                            "read_as_stream": True,
                                            "renames": {"symbol2": "symbol"},
                                            "selects": None,
                                            "spark_chain": None,
                                            "watermark": None,
                                            "catalog_name": "dev",
                                            "cdc": None,
                                            "fmt": "DELTA",
                                            "from_pipeline": True,
                                            "name": "slv_stockmeta",
                                            "path": None,
                                            "schema_name": "markets",
                                        }
                                    },
                                    "on": {"value": ["symbol"]},
                                },
                                "spark_func_name": "smart_join",
                                "sql_expression": None,
                            },
                            {
                                "allow_missing_column_args": False,
                                "column": {
                                    "name": "symbol3",
                                    "type": "string",
                                    "unit": None,
                                },
                                "spark_func_args": [],
                                "spark_func_kwargs": {},
                                "spark_func_name": None,
                                "sql_expression": "symbol",
                            },
                            {
                                "allow_missing_column_args": False,
                                "column": None,
                                "spark_func_args": [{"value": "symbol"}],
                                "spark_func_kwargs": {},
                                "spark_func_name": "drop",
                                "sql_expression": None,
                            },
                            {
                                "allow_missing_column_args": False,
                                "column": None,
                                "spark_func_args": [],
                                "spark_func_kwargs": {
                                    "other": {
                                        "value": {
                                            "drops": None,
                                            "filter": None,
                                            "broadcast": False,
                                            "read_as_stream": True,
                                            "renames": None,
                                            "selects": None,
                                            "spark_chain": None,
                                            "watermark": None,
                                            "catalog_name": "dev",
                                            "cdc": None,
                                            "fmt": "DELTA",
                                            "from_pipeline": True,
                                            "name": "slv_stock_names",
                                            "path": None,
                                            "schema_name": "markets",
                                        }
                                    },
                                    "on": {"value": ["symbol3"]},
                                },
                                "spark_func_name": "smart_join",
                                "sql_expression": None,
                            },
                        ]
                    },
                },
                "catalog_name": "dev1",
                "columns": [],
                "comment": None,
                "data": None,
                "data_source_format": "DELTA",
                "expectations": [],
                "grants": None,
                "name": "slv_join_stock_prices",
                "primary_key": None,
                "schema_name": "markets1",
                "table_type": "MANAGED",
                "timestamp_key": None,
                "view_definition": None,
                "warehouse_id": "08b717ce051a0261",
            },
        ],
        "target": "markets1",
        "udfs": [
            {
                "module_name": "stock_functions",
                "function_name": "high",
                "module_path": None,
            }
        ],
    }


def test_pipeline_pulumi():
    assert pl.resource_name == "pl-stock-prices"
    assert pl.options.model_dump(exclude_none=True) == {
        "depends_on": [],
        "delete_before_replace": True,
    }
    print(pl.pulumi_properties)
    assert pl.pulumi_properties == {
        "catalog": "dev1",
        "channel": "PREVIEW",
        "clusters": [],
        "configuration": {},
        "name": "pl-stock-prices",
        "notifications": [],
        "target": "markets1",
    }

    # Resources
    assert len(pl.core_resources) == 3
    r = pl.core_resources[-1]
    r.options.aliases = ["my-file"]
    assert pl.core_resources[-1].options.aliases == ["my-file"]


if __name__ == "__main__":
    test_pipeline()
    test_pipeline_pulumi()



from abc import abstractmethod
from typing import Any
from typing import Union
from laktory.spark import DataFrame
from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.spark.sparkchain import SparkChain


class Watermark(BaseModel):
    """
    Definition of a spark structured streaming watermark for joining data
    streams.

    Attributes
    ----------
    column:
        Event time column name
    threshold:
        How late, expressed in seconds, the data is expected to be with
        respect to event time.

    References
    ----------
    https://spark.apache.org/docs/latest/structured-streaming-programming-guide.html#handling-late-data-and-watermarking
    """

    column: str
    threshold: str


class BaseDataSource(BaseModel):
    """
    Base class for building data source

    Attributes
    ----------
    drops:
        List of columns to drop
    selects:
        Columns to select from the source table. Can be specified as a list
        or as a dictionary to rename the source columns
    filter:
        SQL expression used to select specific rows from the source table
    read_as_stream
        If `True` read source as stream
    renames:
        Mapping between the source table column names and new column names
    watermark
        Spark structured streaming watermark specifications

    """

    drops: Union[list, None] = None
    filter: Union[str, None] = None
    mock_df: Any = Field(default=None, exclude=True)
    broadcast: Union[bool, None] = False
    read_as_stream: Union[bool, None] = True
    renames: Union[dict[str, str], None] = None
    selects: Union[list[str], dict[str, str], None] = None
    spark_chain: Union[SparkChain, None] = None
    watermark: Union[Watermark, None] = None

    def read(self, spark) -> DataFrame:
        """
        Read data with options specified in attributes.

        Parameters
        ----------
        spark
            Spark context

        Returns
        -------
        : DataFrame
            Resulting park dataframe
        """
        df = self._read(spark)
        df = self._post_read(df)
        return df

    @abstractmethod
    def _read(self, spark) -> DataFrame:
        raise NotImplementedError()

    def _post_read(self, df) -> DataFrame:
        import pyspark.sql.functions as F

        # Apply filter
        if self.filter:
            df = df.filter(self.filter)

        # Apply drops
        if self.drops:
            df = df.drop(*self.drops)

        # Columns
        cols = []
        if self.selects:
            if isinstance(self.selects, list):
                cols += [F.col(c) for c in self.selects]
            elif isinstance(self.selects, dict):
                cols += [F.col(k).alias(v) for k, v in self.selects.items()]
            df = df.select(cols)

        # Renames
        if self.renames:
            for old_name, new_name in self.renames.items():
                df = df.withColumnRenamed(old_name, new_name)

        # Apply Watermark
        if self.watermark:
            df = df.withWatermark(
                self.watermark.column,
                self.watermark.threshold,
            )

        # Broadcast
        if self.broadcast:
            df = F.broadcast(df)

        # SparkChain
        if self.spark_chain:
            df = self.spark_chain.execute(df, udfs=None, spark=df.sparkSession)

        return df

    @property
    def is_cdc(self) -> bool:
        """If `True` source data is a change data capture (CDC)"""
        return getattr(self, "cdc", None) is not None



# MAGIC %pip install laktory

# COMMAND ----------
import pyspark.sql.functions as F
import importlib
import sys
import os

from laktory import dlt
from laktory import read_metadata
from laktory import get_logger

dlt.spark = spark
logger = get_logger(__name__)

# Read pipeline definition
pl_name = spark.conf.get("pipeline_name", "pl-stock-prices")
pl = read_metadata(pipeline=pl_name)

# Import User Defined Functions
sys.path.append("/Workspace/pipelines/")
udfs = []
for udf in pl.udfs:
    if udf.module_path:
        sys.path.append(os.path.abspath(udf.module_path))
    module = importlib.import_module(udf.module_name)
    udfs += [getattr(module, udf.function_name)]


# --------------------------------------------------------------------------- #
# Non-CDC Tables                                                              #
# --------------------------------------------------------------------------- #


def define_table(table):
    @dlt.table_or_view(
        name=table.name,
        comment=table.comment,
        as_view=table.builder.as_dlt_view,
    )
    @dlt.expect_all(table.warning_expectations)
    @dlt.expect_all_or_drop(table.drop_expectations)
    @dlt.expect_all_or_fail(table.fail_expectations)
    def get_df():
        logger.info(f"Building {table.name} table")

        # Read Source
        df = table.builder.read_source(spark)
        df.printSchema()

        # Process
        df = table.builder.process(df, udfs=udfs, spark=spark)

        # Return
        return df

    return get_df


# --------------------------------------------------------------------------- #
# Execution                                                                   #
# --------------------------------------------------------------------------- #

# Build tables
for table in pl.tables:
    if table.builder.template == "SILVER_STAR":
        wrapper = define_table(table)
        df = dlt.get_df(wrapper)
        display(df)
